import re
from collections import namedtuple
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Union

from django.core.exceptions import ValidationError
from django.db.models import Model, QuerySet
from django.utils.functional import cached_property, lazy
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from schwifty.exceptions import InvalidAccountCode
from schwifty.iban import IBAN

from ..conf import settings
from ..utils import currency, paragraph

Price = Union[Decimal, int]


class PaymentStatusMixin:
    price: Price
    discount: Price
    received: Price
    returned: Price
    amount_due: Price
    overdue: Price
    overpaid: Price

    @property
    def paid(self) -> Price:
        return self.received - self.returned

    @property
    def receivable(self) -> Price:
        return self.price - self.discount

    @property
    def balance(self) -> Price:
        return self.paid - self.receivable

    @property
    def color(self) -> str:
        if self.overdue:
            return settings.LEPRIKON_COLOR_OVERDUE
        if self.amount_due:
            return settings.LEPRIKON_COLOR_NOTPAID
        if self.overpaid:
            return settings.LEPRIKON_COLOR_OVERPAID
        return settings.LEPRIKON_COLOR_PAID

    @property
    def amount_due_color(self) -> str:
        if self.overdue:
            return settings.LEPRIKON_COLOR_OVERDUE
        if self.amount_due:
            return settings.LEPRIKON_COLOR_NOTPAID
        return settings.LEPRIKON_COLOR_PAID

    @property
    def overpaid_color(self) -> str:
        if self.overpaid:
            return settings.LEPRIKON_COLOR_OVERPAID
        return settings.LEPRIKON_COLOR_PAID

    @property
    def title(self) -> str:
        titles = []
        if self.amount_due and (self.amount_due != self.overdue):
            titles.append(_("amount due {}").format(currency(self.amount_due)))
        if self.overdue:
            titles.append(_("overdue {}").format(currency(self.overdue)))
        if self.overpaid:
            titles.append(_("{} overpaid").format(currency(self.overpaid)))
        if self.balance == 0:
            if self.receivable:
                titles.append(_("paid"))
            else:
                titles.append(_("payment not requested"))
        elif self.balance < 0 and not self.amount_due:  # self.due_from > self.current_date:
            titles.append(_("payment not requested yet"))
        return ", ".join(map(str, titles))

    def __xrepr__(self) -> str:
        return (
            "{type_name}(price={price}, discount={discount}, received={received}, returned={returned}, "
            "balance={balance}, amount_due={amount_due}, overdue={overdue}, overpaid={overpaid})".format(
                type_name=type(self).__name__,
                price=self.price,
                discount=self.discount,
                received=self.received,
                returned=self.returned,
                balance=self.balance,
                amount_due=self.amount_due,
                overdue=self.overdue,
                overpaid=self.overpaid,
            )
        )

    def __add__(self, other):
        return PaymentStatusSum(
            price=self.price + (other and other.price),
            discount=self.discount + (other and other.discount),
            received=self.received + (other and other.received),
            returned=self.returned + (other and other.returned),
            amount_due=self.amount_due + (other and other.amount_due),
            overdue=self.overdue + (other and other.overdue),
            overpaid=self.overpaid + (other and other.overpaid),
        )

    __radd__ = __add__


class PaymentStatus(
    PaymentStatusMixin,
    namedtuple(
        "_PaymentsStatus",
        ["price", "discount", "explanation", "received", "returned", "current_date", "due_from", "due_date"],
    ),
):
    explanation: str
    current_date: date
    due_from: date
    due_date: date

    @property
    def amount_due(self) -> Price:
        return max(self.receivable - self.paid, 0) if self.due_from and self.due_from <= self.current_date else 0

    @property
    def overdue(self) -> Price:
        return self.amount_due if self.due_date and self.due_date < self.current_date else 0

    @property
    def overpaid(self) -> Price:
        return max(self.balance, 0)

    def __str__(self) -> str:
        return format_html(
            '<strong title="{title}" style="color: {color}">{balance}</strong>',
            color=self.color,
            balance=currency(self.balance),
            title=self.title,
        )


class PaymentStatusSum(
    PaymentStatusMixin,
    namedtuple(
        "_PaymentStatusSum",
        ["price", "discount", "received", "returned", "amount_due", "overdue", "overpaid"],
    ),
):
    pass


class BankAccount(IBAN):
    czech_bban_regex = re.compile(r"(([0-9]{1,6})\s?-\s?)?([0-9]{1,10})\s?/\s?([0-9]{4})")
    czech_bban_weights = (1, 2, 4, 8, 5, 10, 9, 7, 3, 6)

    def _validate_czech(self, part: str) -> bool:
        check = 0
        for i, n in enumerate(reversed(part)):
            check += int(n) * self.czech_bban_weights[i]
        if check % 11:
            raise InvalidAccountCode()

    def __init__(self, bank_account: str):
        match = self.czech_bban_regex.match(bank_account)
        if match:
            _, prefix, account, bank_code = match.groups()
            self._validate_czech(prefix or "")
            self._validate_czech(account)
            account_code = (prefix or "").zfill(6) + account.zfill(10)
            bank_account = f"CZ??{bank_code}{account_code}"
        super().__init__(bank_account)

    @cached_property
    def account_prefix(self):
        return self.branch_code.lstrip("0")

    @cached_property
    def account_number(self):
        return self.account_code.lstrip("0")

    def __str__(self):
        if self.country_code == "CZ":
            return "%s%s%s/%s" % (
                self.account_prefix,
                self.account_prefix and "-",
                self.account_number,
                self.bank_code,
            )
        else:
            return self.formatted


def parse_bank_account(bank_account: Any) -> BankAccount:
    if isinstance(bank_account, BankAccount):
        return bank_account
    if isinstance(bank_account, IBAN):
        return BankAccount(bank_account.compact)
    if not isinstance(bank_account, str):
        bank_account = str(bank_account)
    try:
        return BankAccount(bank_account)
    except ValueError:
        raise ValidationError(
            message=_("Enter a valid account number."),
            code="invalid",
        )


def generate_variable_symbol(registration):
    # get base variable symol from the configured expression
    variable_symbol = eval(settings.LEPRIKON_VARIABLE_SYMBOL_EXPRESSION, {"reg": registration})

    # add check digit
    odd_sum = 0
    even_sum = 0
    for i, char in enumerate(str(variable_symbol)):
        if i % 2:
            even_sum += int(char)
        else:
            odd_sum += int(char)
    check_digit = (odd_sum * 3 + even_sum) % 10
    return variable_symbol * 10 + check_digit


def help_text_with_html_default(help_text, html_default):
    keep_empty = paragraph(_("Keep empty to use default value:"))
    return mark_safe(
        "{}{}{}".format(
            paragraph(help_text),
            keep_empty,
            html_default,
        )
        if help_text
        else "{}{}".format(
            keep_empty,
            html_default,
        )
    )


lazy_help_text_with_html_default = lazy(help_text_with_html_default, str)


def help_text_with_default(help_text, default):
    keep_empty_default = _("Keep empty to use default value: {}").format(default)
    return paragraph("{}\n\n{}".format(help_text, keep_empty_default) if help_text else keep_empty_default)


lazy_help_text_with_default = lazy(help_text_with_default, str)


def change_year(d, year_delta):
    if d is None:
        return None
    if isinstance(d, date):
        try:
            return date(d.year + year_delta, d.month, d.day)
        except ValueError:
            # handle leap-year
            return date(d.year + year_delta, d.month, d.day - 1)
    else:
        try:
            return datetime(d.year + year_delta, d.month, d.day)
        except ValueError:
            # handle leap-year
            return datetime(d.year + year_delta, d.month, d.day - 1, d.hour, d.minute, d.second)


def shorten(string, length):
    if len(string) <= length:
        return string
    half = length // 2 - 1
    return f"{string[:half]}...{string[half + 3 - length:]}"


def copy_related_objects(target: Model, **kwargs: QuerySet):
    for attr, qs in kwargs.items():
        related_objects = list(qs.all())
        remote_field_name = target._meta.get_field(attr).remote_field.name
        for obj in related_objects:
            obj.pk = None
            setattr(obj, remote_field_name, target)
        qs.model.objects.bulk_create(related_objects)



from django.urls import reverse_lazy as reverse
from django.utils.translation import gettext_lazy as _

from ..forms.participant import ParticipantForm
from ..models.roles import Participant
from ..utils import reverse_with_back
from .generic import CreateView, DeleteView, ListView, UpdateView


class ParticipantListView(ListView):
    add_label = _("add participant")
    model = Participant
    template_name = "leprikon/participants.html"
    preview_template = "leprikon/participant_preview.html"

    def get_title(self):
        return _("Participants and parents")

    def get_queryset(self):
        return self.request.user.leprikon_participants.all()

    def get_add_url(self):
        return reverse_with_back(self.request, "leprikon:participant_create")


class ParticipantCreateView(CreateView):
    model = Participant
    form_class = ParticipantForm
    success_url = reverse("leprikon:summary")
    title = _("New participant")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        parent = self.request.user.leprikon_parents.first()
        if parent:
            kwargs["initial"] = dict((attr, getattr(parent, attr)) for attr in ["street", "city", "postal_code"])
        return kwargs

    def get_message(self):
        return _("New participant {} has been created.").format(self.object)


class ParticipantUpdateView(UpdateView):
    model = Participant
    form_class = ParticipantForm
    success_url = reverse("leprikon:summary")
    title = _("Change participant")

    def get_queryset(self):
        # only allow to edit user's own participants
        return super().get_queryset().filter(user=self.request.user)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def get_message(self):
        return _("The participant {} has been updated.").format(self.object)


class ParticipantDeleteView(DeleteView):
    model = Participant
    title = _("Delete information about participant")
    message = _("Information about participant has been deleted.")

    def get_queryset(self):
        # only allow to delete user's own participants
        return super().get_queryset().filter(user=self.request.user)

    def get_question(self):
        return _("Do You really want to delete the information about participant: {}?").format(self.object)



from django.urls import include, path
from drf_spectacular.views import SpectacularAPIView, SpectacularRedocView, SpectacularSwaggerView
from rest_framework.routers import DefaultRouter

from . import views

app_name = "api"

api_router = DefaultRouter(trailing_slash=False)
api_router.register(r"journal", views.JournalViewSet, "journal")
api_router.register(r"schoolyear", views.SchoolYearViewSet, "schoolyear")
api_router.register(r"user", views.UserViewSet, "user")

urlpatterns = [
    path("api/", include(api_router.urls)),
    path("schema/", SpectacularAPIView.as_view(), name="schema"),
    path("swagger/", SpectacularSwaggerView.as_view(url_name="api:schema"), name="swagger"),
    path("docs/", SpectacularRedocView.as_view(url_name="api:schema"), name="redoc"),
]



from django.utils.translation import gettext_lazy as _

from ..models.donation import Donation
from .generic import DetailView, ListView
from .subjects import PDFMixin


class DonationViewMixin:
    model = Donation

    def get_queryset(self):
        return super().get_queryset().filter(donor=self.request.user)


class DonationListView(DonationViewMixin, ListView):
    template_name = "leprikon/donations.html"
    paginate_by = 20

    def get_title(self):
        return _("Donations")


class DonationPdfView(DonationViewMixin, PDFMixin, DetailView):
    pass



# Generated by Django 1.11.13 on 2018-08-26 15:58

from itertools import count
from django.db import migrations, models
import django.db.models.deletion


def create_school_year_divisions(apps, schema_editor):
    Course = apps.get_model('leprikon', 'Course')
    CourseDiscount = apps.get_model('leprikon', 'CourseDiscount')
    SchoolYearDivision = apps.get_model('leprikon', 'SchoolYearDivision')
    SchoolYearPeriod = apps.get_model('leprikon', 'SchoolYearPeriod')
    new_period_ids = {}
    new_period_ids_fuzzy = {}
    school_year_division_ids = {}
    for course in Course.objects.order_by('id').iterator():
        key = tuple(course.periods.order_by('id').values_list('id', flat=True))
        if key not in school_year_division_ids:
            name = course.unit
            counter = count(2)
            while SchoolYearDivision.objects.filter(
                school_year_id=course.school_year_id,
                name=name,
            ).exists():
                name='{course_unit} {c}'.format(course_unit=course.unit, c=next(counter))
            school_year_division = SchoolYearDivision.objects.create(
                school_year_id=course.school_year_id,
                name=name,
                period_name=course.unit,
            )
            for period in course.periods.all():
                old_period_id = period.id
                period.id, period.pk = None, None
                period.school_year_division = school_year_division
                period.save()
                new_period_ids[(key, old_period_id)] = period.id
                new_period_ids_fuzzy[(key, period.name)] = period.id
            school_year_division_ids[key] = school_year_division.id
        course.school_year_division_id = school_year_division_ids[key]
        course.save()
        # fix discounts
        for discount in CourseDiscount.objects.filter(
            registration__subject_id=course.id
        ).select_related('period', 'registration'):
            old_period = discount.period
            discount.period = None
            discount.period_id = new_period_ids.get((key, old_period.id))
            if discount.period_id is None:
                discount.period_id = new_period_ids_fuzzy.get((key, old_period.name))
            discount.save()
    # delete original periods
    SchoolYearPeriod.objects.filter(school_year_division=None).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0014_variable_symbol'),
    ]

    operations = [
        migrations.CreateModel(
            name='SchoolYearDivision',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, verbose_name='division name')),
                ('period_name', models.CharField(max_length=150, verbose_name='period name')),
                ('school_year', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='divisions', to='leprikon.SchoolYear', verbose_name='school year')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'school year division',
                'verbose_name_plural': 'school year divisions',
            },
        ),
        migrations.AlterUniqueTogether(
            name='schoolyeardivision',
            unique_together=set([('school_year', 'name')]),
        ),
        migrations.AddField(
            model_name='course',
            name='school_year_division',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='courses', to='leprikon.SchoolYearDivision', verbose_name='school year division'),
        ),
        migrations.AddField(
            model_name='schoolyearperiod',
            name='school_year_division',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='periods', to='leprikon.SchoolYearDivision', verbose_name='school year division'),
        ),
        migrations.RunPython(create_school_year_divisions),
    ]



from django.contrib import admin

from ..models.question import Question
from .filters import ActiveListFilter


@admin.register(Question)
class QuestionAdmin(admin.ModelAdmin):
    list_display = ("name", "question", "field")
    list_filter = (ActiveListFilter,)
    prepopulated_fields = {"slug": ("name",)}



# Generated by Django 1.11.13 on 2018-08-31 08:43

from django.db import migrations
import djangocms_text_ckeditor.fields


def convert_agreements(apps, schema_editor):
    for model_name in ('LeprikonSite', 'SubjectType', 'Subject'):
        Model = apps.get_model('leprikon', model_name)
        for obj in Model.objects.exclude(agreement=''):
            obj.agreement = '\n'.join(
                '<p>%s</p>' % line for line in obj.agreement.split('\n')
            )
            obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0019_department'),
    ]

    operations = [
        migrations.AlterField(
            model_name='leprikonsite',
            name='agreement',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, default='', verbose_name='registration agreement'),
        ),
        migrations.AlterField(
            model_name='subject',
            name='agreement',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, default='', verbose_name='registration agreement'),
        ),
        migrations.AlterField(
            model_name='subjecttype',
            name='agreement',
            field=djangocms_text_ckeditor.fields.HTMLField(blank=True, default='', verbose_name='registration agreement'),
        ),
        migrations.RunPython(convert_agreements),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-15 06:54
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0026_online_payments'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='coursejournalentry',
            new_name='journalentry',
        ),
        migrations.AlterField(
            model_name='journalentry',
            name='registrations',
            field=models.ManyToManyField(blank=True, related_name='journal_entries', to='leprikon.SubjectRegistration', verbose_name='participants'),
        ),
        migrations.RenameField(
            model_name='journalentry',
            old_name='course',
            new_name='subject',
        ),
        migrations.AlterField(
            model_name='journalentry',
            name='subject',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.PROTECT, related_name='journal_entries', to='leprikon.Subject', verbose_name='subject'),
        ),
        migrations.RenameModel(
            old_name='coursejournalleaderentry',
            new_name='journalleaderentry',
        ),
        migrations.AlterModelOptions(
            name='journalleaderentry',
            options={'verbose_name': 'journal leader entry', 'verbose_name_plural': 'journal leader entries'},
        ),
        migrations.RenameField(
            model_name='journalleaderentry',
            old_name='course_entry',
            new_name='journal_entry',
        ),
        migrations.AlterField(
            model_name='journalleaderentry',
            name='journal_entry',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='leader_entries', to='leprikon.JournalEntry', verbose_name='journal entry'),
        ),
        migrations.AlterField(
            model_name='journalleaderentry',
            name='timesheet',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.PROTECT, related_name='journal_entries', to='leprikon.Timesheet', verbose_name='timesheet'),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from ..forms.billing_info import BillingInfoForm
from ..models.roles import BillingInfo
from ..utils import reverse_with_back
from .generic import CreateView, DeleteView, ListView, UpdateView


class GetQerysetMixin:
    def get_queryset(self):
        return self.request.user.leprikon_billing_info.all()


class BillingInfoListView(GetQerysetMixin, ListView):
    add_label = _("add billing information")
    preview_template = "leprikon/billing_info_preview.html"

    def get_title(self):
        return _("Billing information")

    def get_add_url(self):
        return reverse_with_back(self.request, "leprikon:billing_info_create")


class GetFromKwargsMixin:
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs


class BillingInfoCreateView(GetFromKwargsMixin, GetQerysetMixin, CreateView):
    form_class = BillingInfoForm
    title = _("New billing information")

    def get_message(self):
        return _("New billing information {} have been created.").format(self.object)


class BillingInfoUpdateView(GetFromKwargsMixin, GetQerysetMixin, UpdateView):
    form_class = BillingInfoForm
    message = _("Billing information have been updated.")
    title = _("Change billing information")


class BillingInfoDeleteView(GetQerysetMixin, DeleteView):
    model = BillingInfo
    title = _("Delete billing information")
    message = _("Billing information have been deleted.")

    def get_question(self):
        return _("Do You really want to delete the billing information: {}?").format(self.object)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-10-14 09:31
from __future__ import unicode_literals

from django.db import migrations, models
import re
import django.db.models.deletion


def set_agegroup_age_limits(apps, schema_editor):
    AgeGroup = apps.get_model('leprikon', 'AgeGroup')
    for age_group in AgeGroup.objects.all():
        match = re.match('^[^0-9]*([0-9]+)[^0-9]+([0-9]+)[^0-9]*$', age_group.name)
        if match:
            min_age, max_age = match.groups()
            age_group.min_age = int(min_age)
            age_group.max_age = int(max_age)
            age_group.save()


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0057_registration_period'),
    ]

    operations = [
        migrations.AddField(
            model_name='agegroup',
            name='max_age',
            field=models.PositiveSmallIntegerField(blank=True, null=True, verbose_name='maximal age'),
        ),
        migrations.AddField(
            model_name='agegroup',
            name='min_age',
            field=models.PositiveSmallIntegerField(blank=True, null=True, verbose_name='minimal age'),
        ),
        migrations.RunPython(set_agegroup_age_limits),
    ]



from django.contrib.auth import get_user_model
from rest_framework import serializers

from leprikon.models.schoolyear import SchoolYear
from leprikon.models.subjects import SubjectRegistrationParticipant


class SubjectRegistrationParticipantSerializer(serializers.ModelSerializer):
    class Meta:
        model = SubjectRegistrationParticipant
        fields = ["id", "label"]


class SchoolYearSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = SchoolYear
        fields = ["id", "name", "year", "active"]


class SetSchoolYearSerializer(serializers.Serializer):
    id = serializers.IntegerField()


class CredentialsSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField()


class UserSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = get_user_model()
        fields = ["username", "first_name", "last_name", "email"]



__version__ = "3.5.3"
default_app_config = "leprikon.apps.LeprikonConfig"

staticfiles_urls = {
    "leprikon/lib/bootstrap/css/bootstrap-theme.css": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap-theme.css",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap.min.css": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap.min.css",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap-theme.css.map": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap-theme.css.map",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap.min.css.map": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap.min.css.map",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap-theme.min.css": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap-theme.min.css",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap.css": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap.css",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap-theme.min.css.map": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap-theme.min.css.map",  # NOQA
    "leprikon/lib/bootstrap/css/bootstrap.css.map": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/css/bootstrap.css.map",  # NOQA
    "leprikon/lib/bootstrap/fonts/glyphicons-halflings-regular.svg": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/fonts/glyphicons-halflings-regular.svg",  # NOQA
    "leprikon/lib/bootstrap/fonts/glyphicons-halflings-regular.ttf": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/fonts/glyphicons-halflings-regular.ttf",  # NOQA
    "leprikon/lib/bootstrap/fonts/glyphicons-halflings-regular.eot": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/fonts/glyphicons-halflings-regular.eot",  # NOQA
    "leprikon/lib/bootstrap/fonts/glyphicons-halflings-regular.woff": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/fonts/glyphicons-halflings-regular.woff",  # NOQA
    "leprikon/lib/bootstrap/fonts/glyphicons-halflings-regular.woff2": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/fonts/glyphicons-halflings-regular.woff2",  # NOQA
    "leprikon/lib/bootstrap/js/bootstrap.min.js": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/js/bootstrap.min.js",  # NOQA
    "leprikon/lib/bootstrap/js/npm.js": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/js/npm.js",  # NOQA
    "leprikon/lib/bootstrap/js/bootstrap.js": "https://cdnjs.cloudflare.com/ajax/libs/twitter-bootstrap/3.3.7/js/bootstrap.js",  # NOQA
    "leprikon/lib/bootstrap-datetimepicker.min.css": (
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-datetimepicker/4.17.37/css/bootstrap-datetimepicker.min.css",
        "sha384",
        "/XmhRcB6e51B3NO1S5VVWnNKCDM74NlX0cmxZUwCJPg2P/0ECTkBKnjnzNnF7rX5",
    ),  # NOQA
    "leprikon/lib/bootstrap-datetimepicker.min.js": (
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-datetimepicker/4.17.37/js/bootstrap-datetimepicker.min.js",
        "sha384",
        "V936AaxnYTkq53cO2z8e42PFtCCdZhbRttfhgX8U1K9weNYVdCMbDetUKTtllFdc",
    ),  # NOQA
    "leprikon/lib/bootstrap-multiselect.css": (
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-multiselect/0.9.13/css/bootstrap-multiselect.css",
        "sha384",
        "l1tju+TyzPCuWXt+knuFYVhe2ZnnEm0/t9T8BmqNedtIBYj1fmsYrIAap4MU2amm",
    ),  # NOQA
    "leprikon/lib/bootstrap-multiselect.min.js": (
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-multiselect/0.9.13/js/bootstrap-multiselect.min.js",
        "sha384",
        "2q+7Pk3OSOseTKfNi/t8mZzYRIovdg+nfXYeKmmnivkyD+nxosintmGKX7SWcrOG",
    ),  # NOQA
    "leprikon/lib/ekko-lightbox.min.css": (
        "https://cdnjs.cloudflare.com/ajax/libs/ekko-lightbox/4.0.1/ekko-lightbox.min.css",
        "sha384",
        "EsoR0pjPTZTUy5b8kJf6aTfA7quYluD5mNQ/Lz4hsxU2kXuuc6PlXVfBJCFdTXqD",
    ),  # NOQA
    "leprikon/lib/ekko-lightbox.min.js": (
        "https://cdnjs.cloudflare.com/ajax/libs/ekko-lightbox/4.0.1/ekko-lightbox.min.js",
        "sha384",
        "2aVpK/uKiYlyFXHZ36wvxOPZyWeouEKvbSUlqCrkFhyT80n2aSYtT15Tgdn6g41M",
    ),  # NOQA
    "leprikon/lib/jquery.min.js": (
        "https://cdnjs.cloudflare.com/ajax/libs/jquery/1.12.4/jquery.min.js",
        "sha256",
        "ZosEbRLbNQzLpnKIkEdrPv7lOy9C27hHQ+Xp8a4MxAQ=",
    ),  # NOQA
    "leprikon/lib/moment-with-locales.min.js": (
        "https://cdnjs.cloudflare.com/ajax/libs/moment.js/2.12.0/moment-with-locales.min.js",
        "sha384",
        "V0utPzNT28o7j37Ll4BIKqEVcL1yOE45HoCoaLu/CqF4nFaf2nAyZTHKJW1KdJMH",
    ),  # NOQA
    "leprikon/lib/bootstrap-social.min.css": (
        "https://cdnjs.cloudflare.com/ajax/libs/bootstrap-social/5.1.1/bootstrap-social.min.css",
        "sha384",
        "N7gull5mxE2uLDqZ+kcZolhmuopwLtTcThUsVnUS2LVMQljojN9cfJ9A6fUA63zB",
    ),  # NOQA
    "leprikon/lib/font-awesome/css/font-awesome.css": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.css",  # NOQA
    "leprikon/lib/font-awesome/css/font-awesome.css.map": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.css.map",  # NOQA
    "leprikon/lib/font-awesome/css/font-awesome.min.css": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css",  # NOQA
    "leprikon/lib/font-awesome/fonts/FontAwesome.otf": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/FontAwesome.otf",  # NOQA
    "leprikon/lib/font-awesome/fonts/fontawesome-webfont.eot": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/fontawesome-webfont.eot",  # NOQA
    "leprikon/lib/font-awesome/fonts/fontawesome-webfont.svg": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/fontawesome-webfont.svg",  # NOQA
    "leprikon/lib/font-awesome/fonts/fontawesome-webfont.ttf": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/fontawesome-webfont.ttf",  # NOQA
    "leprikon/lib/font-awesome/fonts/fontawesome-webfont.woff": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/fontawesome-webfont.woff",  # NOQA
    "leprikon/lib/font-awesome/fonts/fontawesome-webfont.woff2": "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/fonts/fontawesome-webfont.woff2",  # NOQA
}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-09-01 22:42
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


def set_registration_questions_agreements(apps, schema_editor):
    leprikon_site = apps.get_model('leprikon', 'LeprikonSite').objects.filter(id=settings.SITE_ID).first()
    site_agreements = set(leprikon_site.registration_agreements.all()) if leprikon_site else set()
    for subject_type in apps.get_model('leprikon', 'SubjectType').objects.all():
        subject_type_questions = set(subject_type.questions.all())
        subject_type_agreements = site_agreements.union(subject_type.registration_agreements.all())
        for subject in subject_type.subjects.prefetch_related('questions', 'registration_agreements'):
            questions = subject_type_questions.union(subject.questions.all())
            agreements = subject_type_agreements.union(subject.registration_agreements.all())
            for registration in subject.registrations.all():
                registration.questions.set(questions)
                registration.agreements.set(agreements)


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0030_agreements'),
    ]

    operations = [
        migrations.AddField(
            model_name='agreement',
            name='active',
            field=models.BooleanField(default=True, verbose_name='active'),
        ),
        migrations.AddField(
            model_name='question',
            name='active',
            field=models.BooleanField(default=True, verbose_name='active'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='agreements',
            field=models.ManyToManyField(editable=False, related_name='registrations', to='leprikon.Agreement'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='questions',
            field=models.ManyToManyField(editable=False, related_name='registrations', to='leprikon.Question'),
        ),
        migrations.RunPython(set_registration_questions_agreements),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-10-20 20:32
from __future__ import unicode_literals

from django.db import migrations, models


def set_participants_instructed(apps, schema_editor):
    Subject = apps.get_model('leprikon', 'Subject')
    for subject in Subject.objects.iterator():
        all_instructed = set()
        for journal_entry in subject.journal_entries.all():
            instructed = {
                participant
                for participant in journal_entry.participants.all()
                if participant not in all_instructed
            }
            journal_entry.participants_instructed.set(instructed)
            all_instructed.update(instructed)


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0035_on_delete'),
    ]

    operations = [
        migrations.AddField(
            model_name='journalentry',
            name='participants_instructed',
            field=models.ManyToManyField(blank=True, related_name='instructed', to='leprikon.SubjectRegistrationParticipant', verbose_name='participants instructed about safety and internal rules'),
        ),
        migrations.RunPython(set_participants_instructed),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-02-09 08:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0065_subject_type_page"),
    ]

    operations = [
        migrations.AlterField(
            model_name="subjecttype",
            name="slug",
            field=models.SlugField(unique=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-11-26 00:05
from __future__ import unicode_literals

from django.db import migrations, models
from django.utils.timezone import now


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0022_bankreader'),
    ]

    operations = [
        migrations.CreateModel(
            name='AccountClosure',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('closure_date', models.DateField(db_index=True, verbose_name='closure date')),
            ],
            options={
                'verbose_name': 'account closure',
                'verbose_name_plural': 'account closures',
                'ordering': ('-closure_date',),
            },
        ),
        migrations.RenameField(
            model_name='coursediscount',
            old_name='created',
            new_name='accounted',
        ),
        migrations.RenameField(
            model_name='eventdiscount',
            old_name='created',
            new_name='accounted',
        ),
        migrations.RenameField(
            model_name='subjectpayment',
            old_name='created',
            new_name='accounted',
        ),
        migrations.AlterModelOptions(
            name='coursediscount',
            options={'ordering': ('accounted',), 'verbose_name': 'course discount', 'verbose_name_plural': 'course discounts'},
        ),
        migrations.AlterModelOptions(
            name='eventdiscount',
            options={'ordering': ('accounted',), 'verbose_name': 'event discount', 'verbose_name_plural': 'event discounts'},
        ),
        migrations.AlterModelOptions(
            name='subjectpayment',
            options={'ordering': ('accounted',), 'verbose_name': 'payment', 'verbose_name_plural': 'payments'},
        ),
        migrations.AlterField(
            model_name='coursediscount',
            name='accounted',
            field=models.DateTimeField(default=now, verbose_name='accounted time'),
        ),
        migrations.AlterField(
            model_name='eventdiscount',
            name='accounted',
            field=models.DateTimeField(default=now, verbose_name='accounted time'),
        ),
        migrations.AlterField(
            model_name='subjectpayment',
            name='accounted',
            field=models.DateTimeField(default=now, verbose_name='accounted time'),
        ),
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from ..utils import first_upper


class TargetGroup(models.Model):
    name = models.CharField(_("name"), max_length=50)
    order = models.IntegerField(_("order"), blank=True, default=0)
    require_school = models.BooleanField(_("require school"), default=True)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("target group")
        verbose_name_plural = _("target groups")

    def __str__(self):
        return self.name

    def title(self):
        return first_upper(self.name)



from datetime import date

from django.db.models import F

from ..models.courses import CourseRegistration
from ..models.events import EventRegistration
from ..models.orderables import OrderableRegistration
from ..models.utils import PaymentStatusSum
from .generic import TemplateView


class SummaryView(TemplateView):
    summary = True
    template_name = "leprikon/summary.html"

    def get_context_data(self, **kwargs):
        payment_status = PaymentStatusSum(0, 0, 0, 0, 0, 0, 0)
        overpaid_registrations = []
        for Registration in (CourseRegistration, EventRegistration, OrderableRegistration):
            for registration in (
                Registration.objects.filter(user=self.request.user)
                .annotate(
                    refund_bank_account=F("refund_request__bank_account"),
                )
                .iterator()
            ):
                payment_status += registration.payment_status
                if registration.payment_status.overpaid:
                    overpaid_registrations.append(registration)
        context = super().get_context_data(**kwargs)
        context["user"] = self.request.user
        context["payment_status"] = payment_status
        context["overpaid_registrations"] = overpaid_registrations
        context["new_messages"] = self.request.user.leprikon_messages.filter(viewed=None)
        return context


class LeaderSummaryView(TemplateView):
    summary = True
    template_name = "leprikon/leader_summary.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["subjects"] = self.request.leader.subjects.filter(school_year=self.request.school_year)
        context["timesheets"] = self.request.leader.timesheets.filter(submitted=False, period__end__lte=date.today())
        return context



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-03-22 08:36
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import leprikon.models.fields
import leprikon.models.subjects


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("leprikon", "0049_target_groups_fix"),
    ]

    operations = [
        migrations.CreateModel(
            name="FilteredOrderableListPlugin",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="leprikon_filteredorderablelistplugin",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
            ],
            bases=("cms.cmsplugin",),
        ),
        migrations.CreateModel(
            name="Orderable",
            fields=[
                (
                    "subject_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="leprikon.Subject",
                    ),
                ),
                ("duration", models.DurationField(help_text="Format: HH:MM:SS", verbose_name="duration")),
                (
                    "due_from_days",
                    models.IntegerField(
                        blank=True,
                        help_text="If set, payment request will be sent this number of days before event date. If not set, payment request will be sent when registration is approved.",
                        null=True,
                        verbose_name="number of days to send the payment request before event date",
                    ),
                ),
                (
                    "due_date_days",
                    models.IntegerField(default=0, verbose_name="number of days to due date before event date"),
                ),
            ],
            options={
                "verbose_name": "orderable event",
                "verbose_name_plural": "orderable events",
                "ordering": ("code", "name"),
            },
            bases=("leprikon.subject",),
        ),
        migrations.CreateModel(
            name="OrderableDiscount",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("accounted", models.DateTimeField(default=django.utils.timezone.now, verbose_name="accounted time")),
                (
                    "amount",
                    leprikon.models.fields.PriceField(
                        decimal_places=0, default=0, max_digits=10, verbose_name="discount"
                    ),
                ),
                (
                    "explanation",
                    models.CharField(blank=True, default="", max_length=250, verbose_name="discount explanation"),
                ),
            ],
            options={
                "verbose_name": "orderable event discount",
                "verbose_name_plural": "orderable event discounts",
                "ordering": ("accounted",),
            },
        ),
        migrations.CreateModel(
            name="OrderableListPlugin",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="leprikon_orderablelistplugin",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "template",
                    models.CharField(
                        choices=[("default", "Default"), ("grouped", "Grouped by event groups")],
                        default="default",
                        help_text="The template used to render plugin.",
                        max_length=100,
                        verbose_name="template",
                    ),
                ),
                (
                    "age_groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Keep empty to skip searching by age groups.",
                        related_name="_orderablelistplugin_age_groups_+",
                        to="leprikon.AgeGroup",
                        verbose_name="age groups",
                    ),
                ),
                (
                    "departments",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Keep empty to skip searching by departments.",
                        related_name="_orderablelistplugin_departments_+",
                        to="leprikon.Department",
                        verbose_name="departments",
                    ),
                ),
            ],
            bases=("cms.cmsplugin",),
        ),
        migrations.CreateModel(
            name="OrderablePlugin",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="leprikon_orderableplugin",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                (
                    "template",
                    models.CharField(
                        choices=[("default", "Default")],
                        default="default",
                        help_text="The template used to render plugin.",
                        max_length=100,
                        verbose_name="template",
                    ),
                ),
                (
                    "event",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="leprikon.Orderable",
                        verbose_name="event",
                    ),
                ),
            ],
            bases=("cms.cmsplugin",),
        ),
        migrations.CreateModel(
            name="OrderableRegistration",
            fields=[
                (
                    "subjectregistration_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="leprikon.SubjectRegistration",
                    ),
                ),
                ("start_date", models.DateField(verbose_name="start date")),
                ("start_time", models.TimeField(blank=True, null=True, verbose_name="start time")),
            ],
            options={
                "verbose_name": "orderable event registration",
                "verbose_name_plural": "orderable event registrations",
            },
            bases=("leprikon.subjectregistration",),
        ),
        migrations.AlterField(
            model_name="subjecttype",
            name="subject_type",
            field=models.CharField(
                choices=[("course", "course"), ("event", "event"), ("orderable", "orderable event")],
                max_length=10,
                verbose_name="subjects",
            ),
        ),
        migrations.AddField(
            model_name="orderablelistplugin",
            name="event_types",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by event types.",
                related_name="_orderablelistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
        migrations.AddField(
            model_name="orderablelistplugin",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by groups.",
                related_name="_orderablelistplugin_groups_+",
                to="leprikon.SubjectGroup",
                verbose_name="event groups",
            ),
        ),
        migrations.AddField(
            model_name="orderablelistplugin",
            name="leaders",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by leaders.",
                related_name="_orderablelistplugin_leaders_+",
                to="leprikon.Leader",
                verbose_name="leaders",
            ),
        ),
        migrations.AddField(
            model_name="orderablelistplugin",
            name="school_year",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="leprikon.SchoolYear",
                verbose_name="school year",
            ),
        ),
        migrations.AddField(
            model_name="orderablelistplugin",
            name="target_groups",
            field=models.ManyToManyField(
                blank=True,
                help_text="Keep empty to skip searching by target groups.",
                related_name="_orderablelistplugin_target_groups_+",
                to="leprikon.TargetGroup",
                verbose_name="target groups",
            ),
        ),
        migrations.AddField(
            model_name="orderablediscount",
            name="registration",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="discounts",
                to="leprikon.OrderableRegistration",
                verbose_name="registration",
            ),
        ),
        migrations.AddField(
            model_name="filteredorderablelistplugin",
            name="event_types",
            field=models.ManyToManyField(
                related_name="_filteredorderablelistplugin_event_types_+",
                to="leprikon.SubjectType",
                verbose_name="event types",
            ),
        ),
        migrations.AddField(
            model_name="filteredorderablelistplugin",
            name="school_year",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="leprikon.SchoolYear",
                verbose_name="school year",
            ),
        ),
    ]



from functools import partial

from django import forms
from django.contrib import admin, messages
from django.forms.models import modelform_factory
from django.forms.widgets import CheckboxSelectMultiple
from django.shortcuts import render
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from sentry_sdk import capture_exception

from ..utils import attributes


class BulkUpdateMixin:
    actions = ("bulk_update",)

    bulk_update_fields = None
    bulk_update_exclude = None
    bulk_update_form = None

    @cached_property
    def BulkUpdateFieldsForm(self):
        return type(
            "BulkUpdateFieldsForm",
            (forms.Form,),
            {
                "fields": forms.MultipleChoiceField(
                    label=_("Fields to update"),
                    choices=[
                        (field_name, field.label)
                        for field_name, field in forms.models.fields_for_model(
                            self.model,
                            fields=self.bulk_update_fields,
                            exclude=self.bulk_update_exclude,
                        ).items()
                    ],
                    widget=CheckboxSelectMultiple,
                ),
            },
        )

    @attributes(short_description=_("Bulk update selected items"))
    def bulk_update(self, request, queryset):
        phase = request.POST.get("phase")
        hidden_fields = []
        # phases: None > 'fields' > 'update'
        if phase is not None:
            form = self.BulkUpdateFieldsForm(request.POST)
            if form.is_valid():
                hidden_fields = [("fields", field) for field in form.cleaned_data["fields"]]
                BulkUpdateForm = modelform_factory(
                    self.model,
                    form=self.bulk_update_form or self.form,
                    fields=form.cleaned_data["fields"],
                    formfield_callback=partial(self.formfield_for_dbfield, request=request),
                )
                if phase == "update":
                    form = BulkUpdateForm(request.POST)
                    if form.is_valid():
                        num_updated = 0
                        for instance in queryset.all():
                            try:
                                instance_form = BulkUpdateForm(request.POST, request.FILES, instance=instance)
                                assert instance_form.is_valid()
                                instance_form.save()
                            except Exception:
                                capture_exception()
                                self.message_user(
                                    request,
                                    _("Failed to update item {}.").format(instance),
                                    messages.ERROR,
                                )
                            else:
                                num_updated += 1
                        self.message_user(request, _("{} items were updated.").format(num_updated))
                        return
                else:
                    form = BulkUpdateForm()
                    phase = "update"
        else:
            form = self.BulkUpdateFieldsForm()
            phase = "fields"
        hidden_fields.append(("phase", phase))
        return render(
            request,
            "leprikon/admin/bulk_update_form.html",
            {
                "title": _("Bulk update"),
                "queryset": queryset,
                "opts": self.model._meta,
                "form": form,
                "media": self.media + form.media,
                "action": "bulk_update",
                "action_checkbox_name": admin.helpers.ACTION_CHECKBOX_NAME,
                "hidden_fields": hidden_fields,
            },
        )



from django.contrib import admin

from ..models.donation import Donation
from .transaction import TransactionBaseAdmin, TransactionTypeListFilter


@admin.register(Donation)
class DonationAdmin(TransactionBaseAdmin):
    exclude = ("target_registration",)
    list_display = (
        "id",
        "accounted_with_by",
        "download_tag",
        "transaction_type_label",
        "amount_html",
        "organization",
        "donor",
        "last_updated_with_by",
        "mail_sent",
        "note",
    )
    list_export = (
        "id",
        "accounted",
        "accounted_by",
        "transaction_type_label",
        "amount",
        "organization",
        "donor__username",
        "donor",
        "last_updated",
        "last_updated_by",
        "mail_sent",
        "note",
    )
    list_filter = (
        ("transaction_type", TransactionTypeListFilter),
        "organization",
        "donor",
    )
    search_fields = (
        "donor__first_name",
        "donor__last_name",
        "donor__email",
    )

    def get_form(self, request, obj=None, **kwargs):
        form = super().get_form(request, obj, **kwargs)
        form.base_fields["donor"].required = True
        form.base_fields["organization"].required = True
        return form



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _

from ..forms.refundrequest import DonationForm, PaymentTransferForm, RefundRequestForm
from ..models.refundrequest import RefundRequest
from ..models.subjects import SubjectRegistration
from .generic import CreateView, DeleteView, UpdateView


class RegistrationMixin:
    def dispatch(self, request, pk, **kwargs):
        self.registration = get_object_or_404(
            SubjectRegistration,
            pk=pk,
            user=request.user,
        ).subjectregistration
        if not self.registration.payment_status.overpaid:
            raise Http404
        return super().dispatch(request, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["registration"] = self.registration
        return kwargs


class RefundRequestCreateView(RegistrationMixin, CreateView):
    form_class = RefundRequestForm
    title = _("New refund request")

    def get_message(self):
        return _("Refund request have been created.")


class RefundRequestQuerysetMixin:
    def get_queryset(self):
        return RefundRequest.objects.filter(registration__user=self.request.user)


class RefundRequestUpdateView(RefundRequestQuerysetMixin, UpdateView):
    form_class = RefundRequestForm
    message = _("Refund request has been updated.")
    title = _("Update refund request")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["registration"] = self.object.registration.subjectregistration
        return kwargs


class RefundRequestDeleteView(RefundRequestQuerysetMixin, DeleteView):
    model = RefundRequest
    title = _("Cancel refund request")
    message = _("Refund request has been canceled.")

    def get_question(self):
        return _("Do You really want to cancel the refund request: {}?").format(self.object)


class PaymentTransferCreateView(RegistrationMixin, CreateView):
    form_class = PaymentTransferForm
    title = _("New payment transfer")

    instructions = _(
        "The payment may only be transferred to a registration, "
        "for which the payment is required. "
        "If You have submitted new registration recently, "
        "You may need to wait until the payment is requested for that registration."
    )

    def get_message(self):
        return _("The payment has been transferred.")


class DonationCreateView(RegistrationMixin, CreateView):
    form_class = DonationForm
    title = _("New donation")

    instructions = _("If You wish, You may transfer the overpaid amount to a donation for our organization.")

    def get_message(self):
        return _("Thank You. Your support is very appreciated.")



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-01-04 16:35
from __future__ import unicode_literals

from django.db import migrations
import django.db.models.deletion
import leprikon.models.fields


def add_subject_type_page(apps, schema_editor):
    SubjectType = apps.get_model("leprikon", "SubjectType")
    Page = apps.get_model("cms", "Page")

    for subject_type in SubjectType.objects.all():
        subject_type.page = Page.objects.filter(
            application_urls="LeprikonSubjectTypeApp",
            application_namespace=subject_type.slug,
            publisher_is_draft=True,
        ).first()
        if subject_type.page:
            subject_type.save()


class Migration(migrations.Migration):

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        ("leprikon", "0064_journals_cleanup"),
    ]

    operations = [
        migrations.AddField(
            model_name="subjecttype",
            name="page",
            field=leprikon.models.fields.UniquePageField(
                blank=True,
                help_text="Select page, where the subjects will be listed. New subject type requireds new page. If You assign an existing page, the current content of the page will be replaced with the new subject list.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="leprikon_subject_type",
                to="cms.Page",
                verbose_name="page",
            ),
            preserve_default=False,
        ),
        migrations.RunPython(add_subject_type_page),
    ]



from collections import namedtuple
from datetime import date, datetime, time, timedelta

from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..utils import attributes, comma_separated
from .fields import DAY_OF_WEEK, DayOfWeekField
from .startend import StartEndMixin


class Time(namedtuple("Time", ("date", "start", "end"))):
    def comparable(self):
        return (self.date, self.start or time(0, 0), self.end or time(0, 0))

    def __lt__(self, other):
        return self.comparable() < other.comparable()


class AbstractTime(StartEndMixin, models.Model):
    day_of_week = DayOfWeekField(_("day of week"))
    start = models.TimeField(_("start time"), blank=True, null=True)
    end = models.TimeField(_("end time"), blank=True, null=True)

    class Meta:
        abstract = True
        app_label = "leprikon"
        ordering = ("day_of_week", "start")
        verbose_name = _("time")
        verbose_name_plural = _("times")

    def __str__(self):
        if self.start is not None and self.end is not None:
            return _("{day}, {start:%H:%M} - {end:%H:%M}").format(
                day=self.day,
                start=self.start,
                end=self.end,
            )
        elif self.start is not None:
            return _("{day}, {start:%H:%M}").format(
                day=self.day,
                start=self.start,
            )
        else:
            return str(self.day)

    @cached_property
    def day(self):
        return DAY_OF_WEEK[self.day_of_week]

    @property
    def delta(self):
        if self.start is not None and self.end is not None:
            d = date(1, 1, 1)
            return datetime.combine(d, self.end) - datetime.combine(d, self.start)
        else:
            return timedelta(0)

    def get_next_time(self, now=None):
        now = now or datetime.now()
        daydelta = (self.day_of_week - now.isoweekday()) % 7
        if daydelta == 0 and (isinstance(now, date) or self.start is None or self.start <= now.time()):
            daydelta = 7
        if isinstance(now, datetime):
            next_date = now.date() + timedelta(daydelta)
        else:
            next_date = now + timedelta(daydelta)
        return Time(
            date=next_date,
            start=self.start,
            end=self.end,
        )


class TimesMixin:
    @cached_property
    def all_times(self):
        return list(self.times.all())

    @attributes(short_description=_("times"))
    def get_times_list(self):
        return comma_separated(self.all_times)

    def get_next_time(self, now=None):
        try:
            return min(t.get_next_time(now) for t in self.all_times)
        except ValueError:
            return None



from collections import namedtuple
from datetime import date, timedelta
from typing import List

from cms.models import CMSPlugin
from django.db import models, transaction
from django.dispatch import receiver
from django.utils.formats import date_format
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from .agegroup import AgeGroup
from .department import Department
from .roles import Leader
from .schoolyear import SchoolYear, SchoolYearDivision, SchoolYearPeriod
from .startend import StartEndMixin
from .subjects import Subject, SubjectDiscount, SubjectGroup, SubjectRegistration, SubjectType, SubjectVariant
from .targetgroup import TargetGroup
from .utils import PaymentStatus, change_year, copy_related_objects


class Course(Subject):
    class Meta:
        app_label = "leprikon"
        ordering = ("code", "name")
        verbose_name = _("course")
        verbose_name_plural = _("courses")

    @property
    def registrations_history_registrations(self):
        return CourseRegistration.objects.filter(course_history__course=self).distinct()

    @property
    def inactive_registrations(self):
        return (
            CourseRegistration.objects.filter(course_history__course=self)
            .exclude(id__in=self.active_registrations.all())
            .distinct()
        )

    def copy_to_school_year(old, school_year: SchoolYear):
        new = Course.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.note = ""
        year_delta = school_year.year - old.school_year.year
        new.save()
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        for old_variant in old.all_variants:
            new_variant = SubjectVariant.objects.get(id=old_variant.id)
            new_variant.id, new_variant.pk = None, None
            new_variant.subject = new
            if old_variant.school_year_division:
                new_variant.school_year_division = SchoolYearDivision.objects.filter(
                    school_year=school_year,
                    name=old_variant.school_year_division.name,
                ).first() or old_variant.school_year_division.copy_to_school_year(school_year)
            new_variant.reg_from = change_year(new_variant.reg_from, year_delta)
            new_variant.reg_to = change_year(new_variant.reg_to, year_delta)
            new_variant.save()
            new_variant.age_groups.set(old_variant.age_groups.all())
            new_variant.target_groups.set(old_variant.target_groups.all())
        copy_related_objects(
            new,
            attachments=old.attachments,
            times=old.times,
        )
        return new


class CourseRegistration(SubjectRegistration):
    subject_type = SubjectType.COURSE

    class Meta:
        app_label = "leprikon"
        verbose_name = _("course registration")
        verbose_name_plural = _("course registrations")

    @cached_property
    def all_registration_periods(self) -> List["CourseRegistrationPeriod"]:
        return list(self.course_registration_periods.select_related("period"))

    def get_period_payment_statuses(self, d=None):
        received = self.get_received(d)
        returned = self.get_returned(d)

        for counter, registration_period in enumerate(self.all_registration_periods, start=1):
            period_payment_status = registration_period.get_payment_status(
                received=received,
                returned=returned,
                last_period=counter == len(self.all_registration_periods),
                d=d,
            )
            yield period_payment_status
            received -= period_payment_status.status.received
            returned = 0  # all returned payments are "spent" to the first period
        if not self.all_registration_periods:
            # create fake registration period to ensure some payment status
            registration_period = CourseRegistrationPeriod(
                registration=self,
                period=SchoolYearPeriod(
                    name=_("no period"),
                ),
            )
            yield CourseRegistrationPeriod.PeriodPaymentStatus(
                period=registration_period.period,
                registration_period=registration_period,
                status=PaymentStatus(
                    price=0,
                    discount=0,
                    explanation="",
                    received=received,
                    returned=returned,
                    current_date=d,
                    due_from=None,
                    due_date=None,
                ),
            )

    def get_payment_status(self, d=None):
        return sum(pps.status for pps in self.get_period_payment_statuses(d))

    @cached_property
    def period_payment_statuses(self):
        return list(self.get_period_payment_statuses())

    @transaction.atomic
    def request_payment(self, payment_requested_by):
        self.course_registration_periods.filter(
            period__due_from__lte=date.today(),
        ).update(payment_requested=True)
        super().request_payment(payment_requested_by)


class CourseRegistrationPeriod(models.Model):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name="course_registration_periods",
        verbose_name=_("registration"),
    )
    period = models.ForeignKey(
        SchoolYearPeriod,
        on_delete=models.PROTECT,
        related_name="course_registration_periods",
        verbose_name=_("period"),
    )
    payment_requested = models.BooleanField(_("payment requested"), default=False)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("registered period")
        verbose_name_plural = _("registered periods")
        ordering = ("period__start",)

    def __str__(self):
        return str(self.period)

    @cached_property
    def all_discounts(self):
        # use cached self.registration.all_discounts instead of
        # return list(self.discounts.all())
        return [discount for discount in self.registration.all_discounts if discount.registration_period_id == self.id]

    class PeriodPaymentStatus(
        namedtuple(
            "PeriodPaymentStatus",
            ["period", "registration_period", "status"],
        )
    ):
        period: SchoolYearPeriod
        registration_period: "CourseRegistrationPeriod"
        status: PaymentStatus

    def get_payment_status(self, received, returned, last_period, d=None) -> PeriodPaymentStatus:
        if d is None:
            d = date.today()
        period_price = self.registration.price * self.period.price_units_count
        discount = sum(discount.amount for discount in self.all_discounts if discount.accounted.date() <= d)
        explanation = ",\n".join(
            discount.explanation.strip()
            for discount in self.all_discounts
            if discount.accounted.date() <= d and discount.explanation.strip()
        )
        return self.PeriodPaymentStatus(
            period=self.period,
            registration_period=self,
            status=PaymentStatus(
                price=period_price,
                discount=discount,
                explanation=explanation,
                received=min(period_price - discount + returned, received) if not last_period else received,
                returned=returned,
                current_date=d,
                due_from=self.registration.payment_requested
                and max(
                    self.period.due_from,
                    self.registration.payment_requested.date(),
                ),
                due_date=self.registration.payment_requested
                and max(
                    self.period.due_date,
                    self.registration.payment_requested.date()
                    + timedelta(days=self.registration.subject.min_due_date_days),
                ),
            ),
        )


class CourseDiscount(SubjectDiscount):
    registration = models.ForeignKey(
        CourseRegistration,
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("registration"),
    )
    registration_period = models.ForeignKey(
        CourseRegistrationPeriod,
        on_delete=models.CASCADE,
        related_name="discounts",
        verbose_name=_("period"),
    )

    @cached_property
    def period(self):
        return self.registration_period.period

    class Meta:
        app_label = "leprikon"
        verbose_name = _("course discount")
        verbose_name_plural = _("course discounts")
        ordering = ("accounted",)


class CourseRegistrationHistory(StartEndMixin, models.Model):
    registration = models.ForeignKey(
        CourseRegistration,
        editable=False,
        on_delete=models.PROTECT,
        related_name="course_history",
        verbose_name=_("course"),
    )
    course = models.ForeignKey(
        Course, editable=False, on_delete=models.PROTECT, related_name="registrations_history", verbose_name=_("course")
    )
    start = models.DateField()
    end = models.DateField(blank=True, null=True)

    class Meta:
        app_label = "leprikon"
        ordering = ("start",)
        verbose_name = _("course registration history")
        verbose_name_plural = _("course registration history")

    def __str__(self):
        return "{course}, {start} - {end}".format(
            course=self.course.name,
            start=date_format(self.start, "SHORT_DATE_FORMAT"),
            end=date_format(self.end, "SHORT_DATE_FORMAT") if self.end else _("now"),
        )

    def save(self, *args, **kwargs):
        if self.id:
            original = self.__class__.objects.get(id=self.id)
            min_journal_date = original.journal_entries.aggregate(models.Min("date"))["date__min"]
            max_journal_date = original.journal_entries.aggregate(models.Max("date"))["date__max"]
            # if journal entry exists, start must be set and lower or equal to min journal date
            if min_journal_date and self.start > min_journal_date:
                self.start = min_journal_date
            # end can not be lower than max journal date
            if self.end and max_journal_date and self.end < max_journal_date:
                self.end = max_journal_date
        super().save(*args, **kwargs)


@receiver(models.signals.post_save, sender=CourseRegistration)
def update_course_registration_history(sender, instance, created, **kwargs):
    if instance.approved:
        d = date.today()
        # if created or changed
        if instance.course_history.count() == 0 or instance.course_history.filter(end=None).exclude(
            course_id=instance.subject_id
        ).update(end=d):
            # reopen or create entry starting today
            (
                CourseRegistrationHistory.objects.filter(
                    registration_id=instance.id,
                    course_id=instance.subject_id,
                    start=d,
                ).update(end=None)
                or CourseRegistrationHistory.objects.create(
                    registration_id=instance.id,
                    course_id=instance.subject_id,
                    start=d,
                )
            )


class CoursePlugin(CMSPlugin):
    course = models.ForeignKey(Course, on_delete=models.CASCADE, related_name="+", verbose_name=_("course"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_COURSE_TEMPLATES,
        default=settings.LEPRIKON_COURSE_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class CourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    departments = models.ManyToManyField(
        Department,
        blank=True,
        related_name="+",
        verbose_name=_("departments"),
        help_text=_("Keep empty to skip searching by departments."),
    )
    course_types = models.ManyToManyField(
        SubjectType,
        blank=True,
        limit_choices_to={"subject_type": SubjectType.COURSE},
        related_name="+",
        verbose_name=_("course types"),
        help_text=_("Keep empty to skip searching by course types."),
    )
    age_groups = models.ManyToManyField(
        AgeGroup,
        blank=True,
        related_name="+",
        verbose_name=_("age groups"),
        help_text=_("Keep empty to skip searching by age groups."),
    )
    target_groups = models.ManyToManyField(
        TargetGroup,
        blank=True,
        related_name="+",
        verbose_name=_("target groups"),
        help_text=_("Keep empty to skip searching by target groups."),
    )
    groups = models.ManyToManyField(
        SubjectGroup,
        blank=True,
        related_name="+",
        verbose_name=_("course groups"),
        help_text=_("Keep empty to skip searching by groups."),
    )
    leaders = models.ManyToManyField(
        Leader,
        blank=True,
        related_name="+",
        verbose_name=_("leaders"),
        help_text=_("Keep empty to skip searching by leaders."),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_COURSELIST_TEMPLATES,
        default=settings.LEPRIKON_COURSELIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.course_types.set(oldinstance.course_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.target_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple("Group", ("group", "objects"))

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        courses = Course.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            courses = courses.filter(department__in=self.all_departments)
        if self.all_course_types:
            courses = courses.filter(subject_type__in=self.all_course_types)
        if self.all_age_groups:
            courses = courses.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            courses = courses.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            courses = courses.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            courses = courses.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_course_types:
            groups = SubjectGroup.objects.filter(subject_types__in=self.all_course_types)
        else:
            groups = SubjectGroup.objects.all()

        context.update(
            {
                "school_year": school_year,
                "courses": courses,
                "groups": (self.Group(group=group, objects=courses.filter(groups=group)) for group in groups),
            }
        )
        return context


class FilteredCourseListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    course_types = models.ManyToManyField(
        SubjectType,
        limit_choices_to={"subject_type": SubjectType.COURSE},
        related_name="+",
        verbose_name=_("course types"),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.course_types.set(oldinstance.course_types.all())

    @cached_property
    def all_course_types(self):
        return list(self.course_types.all())

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )

        from ..forms.subjects import SubjectFilterForm

        form = SubjectFilterForm(
            subject_type_type=SubjectType.COURSE,
            subject_types=self.all_course_types,
            school_year=school_year,
            is_staff=context["request"].user.is_staff,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "courses": form.get_queryset(),
            }
        )
        return context



from datetime import date

from cms.models import CMSPlugin
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..utils import comma_separated
from .startend import StartEndMixin
from .utils import change_year


class SchoolYearManager(models.Manager):
    def get_current(self):
        # by default use last active year
        school_year = self.filter(active=True).order_by("-year").first()
        if school_year is None:
            # Create or activate current year
            if date.today().month < 7:
                year = date.today().year - 1
            else:
                year = date.today().year
            school_year = SchoolYear.objects.get_or_create(year=year)[0]
            school_year.active = True
            school_year.save()
        return school_year


class SchoolYear(models.Model):
    year = models.IntegerField(_("year"), unique=True, help_text=_("Only the first of the two years."))
    active = models.BooleanField(_("active"), default=False)

    objects = SchoolYearManager()

    class Meta:
        app_label = "leprikon"
        ordering = ("-year",)
        verbose_name = _("school year")
        verbose_name_plural = _("school years")

    def __str__(self):
        return self.name

    def delete(self):
        if self.courses.count() or self.events.count():
            raise Exception(_("Can not delete sclool year with courses or events."))
        super().delete()

    @cached_property
    def name(self):
        return "{}/{}".format(self.year, self.year + 1)


class SchoolYearDivision(models.Model):
    school_year = models.ForeignKey(
        SchoolYear, on_delete=models.CASCADE, related_name="divisions", verbose_name=_("school year")
    )
    name = models.CharField(_("division name"), max_length=150)
    price_unit_name = models.CharField(
        _("price unit name"),
        max_length=150,
        help_text=_("Unit name the price applies to. For example semester, month, or 5 lessons."),
        # "Název jednotky, za kterou je udávána cena, například pololetí, měsíc, 5 lekcí, apod."
    )

    class Meta:
        app_label = "leprikon"
        ordering = ("name",)
        unique_together = (("school_year", "name"),)
        verbose_name = _("school year division")
        verbose_name_plural = _("school year divisions")

    def __str__(self):
        return self.name

    @cached_property
    def all_periods(self):
        return list(self.periods.all())

    def copy_to_school_year(old, school_year):
        new = SchoolYearDivision.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.save()
        year_delta = school_year.year - old.school_year.year
        periods = []
        for period in old.all_periods:
            periods.append(
                SchoolYearPeriod(
                    school_year_division=new,
                    name=period.name,
                    start=change_year(period.start, year_delta),
                    end=change_year(period.end, year_delta),
                    due_from=change_year(period.due_from, year_delta),
                    due_date=change_year(period.due_date, year_delta),
                )
            )
        SchoolYearPeriod.objects.bulk_create(periods)
        return new

    def get_current_period(self):
        return self.periods.filter(end__gte=date.today()).first() or self.periods.last()


class SchoolYearPeriod(StartEndMixin, models.Model):
    school_year_division = models.ForeignKey(
        SchoolYearDivision, on_delete=models.CASCADE, related_name="periods", verbose_name=_("school year division")
    )
    name = models.CharField(_("name"), max_length=150)
    start = models.DateField(_("start date"), blank=True, null=True)
    end = models.DateField(_("end date"), blank=True, null=True)
    price_units_count = models.DecimalField(_("number of price units"), default=1, decimal_places=2, max_digits=10)
    due_from = models.DateField(_("due from"))
    due_date = models.DateField(_("due date"))
    order = models.IntegerField(_("order"), blank=True, default=0)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("school year period")
        verbose_name_plural = _("school year periods")

    def __str__(self):
        return self.name


class SchoolYearBlockPlugin(CMSPlugin):
    school_years = models.ManyToManyField(
        SchoolYear,
        related_name="+",
        verbose_name=_("school years"),
        help_text=_("Which school years should the block be visible for?"),
    )

    class Meta:
        app_label = "leprikon"

    def __str__(self):
        return comma_separated(self.all_school_years)

    @cached_property
    def all_school_years(self):
        return list(self.school_years.all())

    def copy_relations(self, oldinstance):
        self.school_years.set(oldinstance.school_years.all())



from django import forms


class RegistrationLinkAdminForm(forms.ModelForm):
    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        # limit choices of subjects
        subjects_choices = self.fields["subjects"].widget.choices
        subjects_choices.queryset = subjects_choices.queryset.filter(
            school_year=self.school_year,
            subject_type=self.subject_type,
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-06-21 12:50
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


def set_cancelation_requested(apps, schema_editor):
    SubjectRegistration = apps.get_model('leprikon', 'SubjectRegistration')
    SubjectRegistration.objects.filter(cancel_request=True).update(cancelation_requested='1970-01-01 00:00')


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('leprikon', '0052_payment_notification_sent'),
    ]

    operations = [
        migrations.AddField(
            model_name='subjectregistration',
            name='approved_by',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='approved by'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='cancelation_requested',
            field=models.DateTimeField(editable=False, null=True, verbose_name='time of cancellation request'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='cancelation_requested_by',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='cancelation requested by'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='canceled_by',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='canceled by'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='created_by',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='created by'),
        ),
        migrations.AddField(
            model_name='subjectregistration',
            name='payment_requested_by',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='payment requested by'),
        ),
        migrations.RunPython(set_cancelation_requested),
    ]



from json import loads

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..utils import first_upper


class Question(models.Model):
    name = models.CharField(_("name"), max_length=150)
    slug = models.SlugField(_("unique identifier"), max_length=50, unique=True)
    question = models.CharField(_("question"), max_length=150)
    help_text = models.TextField(
        _("help text"),
        blank=True,
        null=True,
        help_text=_("This is help text. The help text is shown next to the form field."),
    )
    field = models.CharField(
        _("field"),
        max_length=150,
        choices=((key, val["name"]) for key, val in settings.LEPRIKON_QUESTION_FIELDS.items()),
    )
    field_args = models.TextField(
        _("field_args"),
        blank=True,
        default="{}",
        help_text=_("Enter valid JSON structure representing field configuration."),
    )
    active = models.BooleanField(_("active"), default=True)

    class Meta:
        app_label = "leprikon"
        ordering = ("name",)
        verbose_name = _("additional question")
        verbose_name_plural = _("additional questions")

    def __str__(self):
        return self.name

    @cached_property
    def field_class(self):
        return import_string(settings.LEPRIKON_QUESTION_FIELDS[self.field]["class"])

    @cached_property
    def field_kwargs(self):
        return loads(self.field_args)

    @cached_property
    def field_label(self):
        return first_upper(self.question)

    def get_field(self, initial=None):
        return self.field_class(label=self.field_label, initial=initial, help_text=self.help_text, **self.field_kwargs)

    def clean(self):
        try:
            self.get_field()
        except Exception as e:
            raise ValidationError({"field_args": [_("Failed to create field with given field args: {}").format(e)]})



from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from djangocms_text_ckeditor.fields import HTMLField


class Agreement(models.Model):
    name = models.CharField(_("name"), max_length=50, unique=True)
    order = models.IntegerField(_("order"), blank=True, default=0)
    heading = models.CharField(_("heading"), max_length=50, blank=True, default="")
    agreement = HTMLField(_("agreement"), blank=True, default="")
    active = models.BooleanField(_("active"), default=True)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("legal agreement")
        verbose_name_plural = _("legal agreements")

    def __str__(self):
        return self.name

    @cached_property
    def all_options(self):
        return list(self.options.all())


class AgreementOption(models.Model):
    agreement = models.ForeignKey(
        Agreement, on_delete=models.CASCADE, related_name="options", verbose_name=_("agreement")
    )
    name = models.CharField(_("name"), max_length=50)
    order = models.IntegerField(_("order"), blank=True, default=0)
    required = models.BooleanField(_("required"), default=False)
    option = HTMLField(_("option"))

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        unique_together = (("agreement", "name"),)
        verbose_name = _("agreement option")
        verbose_name_plural = _("agreement options")

    def __str__(self):
        return self.name



from . import (  # noqa
    account,
    agegroup,
    agreements,
    citizenship,
    courses,
    department,
    donation,
    events,
    fields,
    journals,
    leprikonsite,
    messages,
    orderables,
    organizations,
    place,
    printsetup,
    question,
    refundrequest,
    registrationlink,
    roles,
    school,
    schoolyear,
    startend,
    subjects,
    targetgroup,
    timesheets,
    transaction,
    useragreement,
    utils,
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2021-06-08 20:40
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import djangocms_text_ckeditor.fields
import multiselectfield.db.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("leprikon", "0061_refundrequest"),
    ]

    operations = [
        migrations.AddField(
            model_name="subject",
            name="text_registration_refund_offer",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True, default="", verbose_name="text: registration refund offer"
            ),
        ),
        migrations.AddField(
            model_name="subjectregistration",
            name="refund_offered",
            field=models.DateTimeField(editable=False, null=True, verbose_name="payment request time"),
        ),
        migrations.AddField(
            model_name="subjectregistration",
            name="refund_offered_by",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to=settings.AUTH_USER_MODEL,
                verbose_name="payment requested by",
            ),
        ),
        migrations.AddField(
            model_name="subjecttype",
            name="text_registration_refund_offer",
            field=djangocms_text_ckeditor.fields.HTMLField(
                blank=True,
                default="",
                help_text="<p>Keep empty to use default value:</p><p>Hello,<br/>\nYour registration has been overpaid.<br/>\nPlease, tell us how you wish us to refund.</p>",
                verbose_name="text: registration refund offer",
            ),
        ),
        migrations.AlterField(
            model_name="subjectattachment",
            name="events",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("registration_received", "registration received"),
                    ("registration_approved", "registration approved"),
                    ("registration_refused", "registration refused"),
                    ("registration_payment_request", "payment requested"),
                    ("registration_refund_offer", "payment requested"),
                    ("registration_canceled", "registration canceled"),
                    ("discount_granted", "discount granted"),
                    ("payment_received", "payment received"),
                    ("payment_returned", "payment returned"),
                ],
                default=[],
                help_text="The attachment will be sent with notification on selected events.",
                max_length=192,
                verbose_name="send when",
            ),
        ),
        migrations.AlterField(
            model_name="subjecttypeattachment",
            name="events",
            field=multiselectfield.db.fields.MultiSelectField(
                blank=True,
                choices=[
                    ("registration_received", "registration received"),
                    ("registration_approved", "registration approved"),
                    ("registration_refused", "registration refused"),
                    ("registration_payment_request", "payment requested"),
                    ("registration_refund_offer", "payment requested"),
                    ("registration_canceled", "registration canceled"),
                    ("discount_granted", "discount granted"),
                    ("payment_received", "payment received"),
                    ("payment_returned", "payment returned"),
                ],
                default=[],
                help_text="The attachment will be sent with notification on selected events.",
                max_length=192,
                verbose_name="send when",
            ),
        ),
    ]



# Generated by Django 3.2.20 on 2023-07-13 14:29

from django.db import migrations, models
import django.db.models.deletion


def copy_course_variant_values(apps, schema_editor):
    CourseVariant = apps.get_model("leprikon", "CourseVariant")

    for cv in CourseVariant.objects.all():
        cv.school_year_division_id = cv.school_year_division_old_id
        cv.allow_period_selection = cv.allow_period_selection_old
        cv.save()


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0077_place_description"),
    ]

    operations = [
        migrations.RenameField(
            model_name="coursevariant",
            old_name="allow_period_selection",
            new_name="allow_period_selection_old",
        ),
        migrations.RenameField(
            model_name="coursevariant",
            old_name="school_year_division",
            new_name="school_year_division_old",
        ),
        migrations.AddField(
            model_name="subjectvariant",
            name="allow_period_selection",
            field=models.BooleanField(
                default=False,
                help_text="allow user to choose school year periods on registration form",
                verbose_name="allow period selection",
            ),
        ),
        migrations.AddField(
            model_name="subjectvariant",
            name="school_year_division",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="variants",
                to="leprikon.schoolyeardivision",
                verbose_name="school year division",
            ),
        ),
        migrations.RunPython(copy_course_variant_values),
        migrations.DeleteModel(name="CourseVariant"),
    ]



from collections import namedtuple
from functools import lru_cache

from django.db.models import Sum
from django.template.response import TemplateResponse
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ...forms.reports.orderables import OrderablePaymentsForm, OrderablePaymentsStatusForm, OrderableStatsForm
from ...models.agegroup import AgeGroup
from ...models.citizenship import Citizenship
from ...models.orderables import Orderable, OrderableRegistration
from ...models.roles import Participant
from ...models.subjects import SubjectPayment, SubjectRegistrationParticipant, SubjectType
from ...views.generic import FormView


class ReportOrderablePaymentsView(FormView):
    form_class = OrderablePaymentsForm
    template_name = "leprikon/reports/orderable_payments.html"
    title = _("Orderable payments")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["received_payments"] = SubjectPayment.objects.filter(
            target_registration__subject__subject_type__subject_type=SubjectType.ORDERABLE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["returned_payments"] = SubjectPayment.objects.filter(
            source_registration__subject__subject_type__subject_type=SubjectType.ORDERABLE,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["received_payments_sum"] = context["received_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["returned_payments_sum"] = context["returned_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["sum"] = context["received_payments_sum"] - context["returned_payments_sum"]
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))


class ReportOrderablePaymentsStatusView(FormView):
    form_class = OrderablePaymentsStatusForm
    template_name = "leprikon/reports/orderable_payments_status.html"
    title = _("Orderable event payments status")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    OrderablePaymentsStatusSums = namedtuple("OrderablePaymentsStatusSums", ("registrations", "status"))

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["reports"] = [
            self.Report(orderable, context["date"])
            for orderable in Orderable.objects.filter(school_year=self.request.school_year)
        ]
        context["sum"] = self.OrderablePaymentsStatusSums(
            registrations=sum(len(r.registration_statuses) for r in context["reports"]),
            status=sum(r.status for r in context["reports"]),
        )
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))

    class Report:
        def __init__(self, orderable, d):
            self.orderable = orderable
            self.date = d

        RegPaymentStatus = namedtuple("RegPaymentStatus", ("registration", "status"))

        @cached_property
        def registration_statuses(self):
            return [
                registration_status
                for registration_status in (
                    self.RegPaymentStatus(
                        registration=registration,
                        status=registration.get_payment_status(self.date),
                    )
                    for registration in OrderableRegistration.objects.filter(
                        subject=self.orderable,
                        approved__date__lte=self.date,
                    )
                )
                if registration_status.status.receivable
            ]

        @cached_property
        def status(self):
            return sum(rs.status for rs in self.registration_statuses)


class ReportOrderableStatsView(FormView):
    form_class = OrderableStatsForm
    template_name = "leprikon/reports/orderable_stats.html"
    title = _("Orderable statistics")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    ReportItem = namedtuple("ReportItem", ("age_group", "all", "boys", "girls", "citizenships"))

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["school_year"] = self.request.school_year
        return kwargs

    @lru_cache
    def get_orderable_days(self, orderable_id):
        orderable = Orderable.objects.get(id=orderable_id)
        return orderable.duration.days + 1

    def form_valid(self, form):
        d = form.cleaned_data["date"]
        paid_only = form.cleaned_data["paid_only"]
        paid_later = form.cleaned_data["paid_later"]
        approved_later = form.cleaned_data["approved_later"]
        unique_participants = form.cleaned_data["unique_participants"]
        min_days = form.cleaned_data["min_days"]
        context = form.cleaned_data
        context["form"] = form

        if approved_later:
            # approved registrations created by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__created__date__lte=d,
                registration__approved__isnull=False,
            )
        else:
            # registrations approved by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__approved__date__lte=d,
            )
        participants = (
            participants.filter(
                registration__subject__in=form.cleaned_data["orderables"],
            )
            .exclude(registration__canceled__date__lte=d)
            .select_related("registration", "age_group")
        )
        if paid_only:
            paid_date = None if paid_later else d
            participants = [
                participant
                for participant in participants
                if participant.registration.orderableregistration.get_payment_status(paid_date).balance >= 0
            ]
        else:
            participants = list(participants)

        if min_days:
            participants = [p for p in participants if self.get_orderable_days(p.registration.subject_id) >= min_days]

        context["orderables_count"] = len(set(participant.registration.subject_id for participant in participants))

        if unique_participants:
            participants = list({p.key: p for p in participants}.values())

        citizenships = list(Citizenship.objects.all())
        context["citizenships"] = citizenships

        context["participants_counts"] = self.ReportItem(
            age_group=None,
            all=len(participants),
            boys=len([p for p in participants if p.gender == Participant.MALE]),
            girls=len([p for p in participants if p.gender == Participant.FEMALE]),
            citizenships=[
                len([p for p in participants if p.citizenship_id == citizenship.id]) for citizenship in citizenships
            ],
        )
        context["participants_counts_by_age_groups"] = []
        for age_group in AgeGroup.objects.all():
            parts = [p for p in participants if p.age_group == age_group]
            context["participants_counts_by_age_groups"].append(
                self.ReportItem(
                    age_group=age_group,
                    all=len(parts),
                    boys=len([p for p in parts if p.gender == Participant.MALE]),
                    girls=len([p for p in parts if p.gender == Participant.FEMALE]),
                    citizenships=[
                        len([p for p in parts if p.citizenship_id == citizenship.id]) for citizenship in citizenships
                    ],
                )
            )

        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))



from collections import namedtuple
from datetime import date, datetime, timedelta

from cms.models import CMSPlugin
from django.db import models
from django.utils.formats import date_format, time_format
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..utils import attributes
from .agegroup import AgeGroup
from .department import Department
from .roles import Leader
from .schoolyear import SchoolYear, SchoolYearDivision
from .subjects import Subject, SubjectDiscount, SubjectGroup, SubjectRegistration, SubjectType, SubjectVariant
from .targetgroup import TargetGroup
from .times import Time
from .utils import PaymentStatus, change_year, copy_related_objects


class Event(Subject):
    start_date = models.DateField(_("start date"))
    end_date = models.DateField(_("end date"))
    start_time = models.TimeField(_("start time"), blank=True, null=True)
    end_time = models.TimeField(_("end time"), blank=True, null=True)
    due_from = models.DateField(_("due from"))
    due_date = models.DateField(_("due date"))

    class Meta:
        app_label = "leprikon"
        ordering = ("start_date", "start_time")
        verbose_name = _("event")
        verbose_name_plural = _("events")

    @property
    def inactive_registrations(self):
        return self.registrations.filter(canceled__isnull=False)

    @attributes(short_description=_("times"))
    def event_date(self):
        return "{start}{separator}{end}".format(
            start=(
                date_format(datetime.combine(self.start_date, self.start_time), "SHORT_DATETIME_FORMAT")
                if self.start_time
                else date_format(self.start_date, "SHORT_DATE_FORMAT")
            ),
            separator=" - " if self.start_date != self.end_date or self.end_time is not None else "",
            end=(
                (time_format(self.end_time, "TIME_FORMAT") if self.end_time else "")
                if self.start_date == self.end_date
                else (
                    date_format(datetime.combine(self.end_date, self.end_time), "SHORT_DATETIME_FORMAT")
                    if self.end_time
                    else date_format(self.end_date, "SHORT_DATE_FORMAT")
                )
            ),
        )

    def get_next_time(self, now=None):
        if not now:
            return Time(
                date=self.start_date,
                start=self.start_time,
                end=(self.end_time if self.end_date == self.start_date else None),
            )
        else:
            next_date = (now.date() if isinstance(now, datetime) else now) + timedelta(1)
            return Time(
                date=next_date,
                start=None,
                end=(self.end_time if self.end_date == next_date else None),
            )

    def copy_to_school_year(old, school_year: SchoolYear):
        new = Event.objects.get(id=old.id)
        new.id, new.pk = None, None
        new.school_year = school_year
        new.public = False
        new.note = ""
        year_delta = school_year.year - old.school_year.year
        new.due_from = change_year(new.due_from, year_delta)
        new.due_date = change_year(new.due_date, year_delta)
        new.start_date = change_year(new.start_date, year_delta)
        new.end_date = change_year(new.end_date, year_delta)
        new.save()
        new.groups.set(old.groups.all())
        new.age_groups.set(old.age_groups.all())
        new.target_groups.set(old.target_groups.all())
        for leader in old.all_leaders:
            school_year.leaders.add(leader)
        new.leaders.set(old.all_leaders)
        new.questions.set(old.questions.all())
        for old_variant in old.all_variants:
            new_variant = SubjectVariant.objects.get(id=old_variant.id)
            new_variant.id, new_variant.pk = None, None
            new_variant.subject = new
            if old_variant.school_year_division:
                new_variant.school_year_division = SchoolYearDivision.objects.filter(
                    school_year=school_year,
                    name=old_variant.school_year_division.name,
                ).first() or old_variant.school_year_division.copy_to_school_year(school_year)
            new_variant.reg_from = change_year(new_variant.reg_from, year_delta)
            new_variant.reg_to = change_year(new_variant.reg_to, year_delta)
            new_variant.save()
            new_variant.age_groups.set(old_variant.age_groups.all())
            new_variant.target_groups.set(old_variant.target_groups.all())
        copy_related_objects(
            new,
            attachments=old.attachments,
            times=old.times,
        )
        return new


class EventRegistration(SubjectRegistration):
    subject_type = SubjectType.EVENT

    class Meta:
        app_label = "leprikon"
        verbose_name = _("event registration")
        verbose_name_plural = _("event registrations")

    def get_payment_status(self, d=None):
        return PaymentStatus(
            price=self.price,
            discount=self.get_discounted(d),
            explanation=",\n".join(
                discount.explanation.strip()
                for discount in self.all_discounts
                if (d is None or discount.accounted.date() <= d) and discount.explanation.strip()
            ),
            received=self.get_received(d),
            returned=self.get_returned(d),
            current_date=d or date.today(),
            due_from=self.payment_requested
            and max(
                self.subject.event.due_from,
                self.payment_requested.date(),
            ),
            due_date=self.payment_requested
            and max(
                self.subject.event.due_date,
                self.payment_requested.date() + timedelta(days=self.subject.min_due_date_days),
            ),
        )


class EventDiscount(SubjectDiscount):
    registration = models.ForeignKey(
        EventRegistration, on_delete=models.CASCADE, related_name="discounts", verbose_name=_("registration")
    )

    class Meta:
        app_label = "leprikon"
        verbose_name = _("event discount")
        verbose_name_plural = _("event discounts")
        ordering = ("accounted",)


class EventPlugin(CMSPlugin):
    event = models.ForeignKey(Event, on_delete=models.CASCADE, related_name="+", verbose_name=_("event"))
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_EVENT_TEMPLATES,
        default=settings.LEPRIKON_EVENT_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"


class EventListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    departments = models.ManyToManyField(
        Department,
        blank=True,
        related_name="+",
        verbose_name=_("departments"),
        help_text=_("Keep empty to skip searching by departments."),
    )
    event_types = models.ManyToManyField(
        SubjectType,
        blank=True,
        limit_choices_to={"subject_type": SubjectType.EVENT},
        related_name="+",
        verbose_name=_("event types"),
        help_text=_("Keep empty to skip searching by event types."),
    )
    age_groups = models.ManyToManyField(
        AgeGroup,
        blank=True,
        related_name="+",
        verbose_name=_("age groups"),
        help_text=_("Keep empty to skip searching by age groups."),
    )
    target_groups = models.ManyToManyField(
        TargetGroup,
        blank=True,
        related_name="+",
        verbose_name=_("target groups"),
        help_text=_("Keep empty to skip searching by target groups."),
    )
    groups = models.ManyToManyField(
        SubjectGroup,
        blank=True,
        related_name="+",
        verbose_name=_("event groups"),
        help_text=_("Keep empty to skip searching by groups."),
    )
    leaders = models.ManyToManyField(
        Leader,
        verbose_name=_("leaders"),
        blank=True,
        related_name="+",
        help_text=_("Keep empty to skip searching by leaders."),
    )
    template = models.CharField(
        _("template"),
        max_length=100,
        choices=settings.LEPRIKON_EVENTLIST_TEMPLATES,
        default=settings.LEPRIKON_EVENTLIST_TEMPLATES[0][0],
        help_text=_("The template used to render plugin."),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.departments.set(oldinstance.departments.all())
        self.event_types.set(oldinstance.event_types.all())
        self.groups.set(oldinstance.groups.all())
        self.age_groups.set(oldinstance.age_groups.all())
        self.target_groups.set(oldinstance.target_groups.all())
        self.leaders.set(oldinstance.leaders.all())

    @cached_property
    def all_departments(self):
        return list(self.departments.all())

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    @cached_property
    def all_groups(self):
        return list(self.groups.all())

    @cached_property
    def all_age_groups(self):
        return list(self.age_groups.all())

    @cached_property
    def all_target_groups(self):
        return list(self.target_groups.all())

    @cached_property
    def all_leaders(self):
        return list(self.leaders.all())

    Group = namedtuple("Group", ("group", "objects"))

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )
        events = Event.objects.filter(school_year=school_year, public=True).distinct()

        if self.all_departments:
            events = events.filter(department__in=self.all_departments)
        if self.all_event_types:
            events = events.filter(subject_type__in=self.all_event_types)
        if self.all_age_groups:
            events = events.filter(age_groups__in=self.all_age_groups)
        if self.all_target_groups:
            events = events.filter(target_groups__in=self.all_target_groups)
        if self.all_leaders:
            events = events.filter(leaders__in=self.all_leaders)
        if self.all_groups:
            events = events.filter(groups__in=self.all_groups)
            groups = self.all_groups
        elif self.all_event_types:
            groups = SubjectGroup.objects.filter(subject_types__in=self.all_event_types)
        else:
            groups = SubjectGroup.objects.all()

        context.update(
            {
                "school_year": school_year,
                "events": events,
                "groups": (self.Group(group=group, objects=events.filter(groups=group)) for group in groups),
            }
        )
        return context


class FilteredEventListPlugin(CMSPlugin):
    school_year = models.ForeignKey(
        SchoolYear, blank=True, null=True, on_delete=models.CASCADE, related_name="+", verbose_name=_("school year")
    )
    event_types = models.ManyToManyField(
        SubjectType,
        limit_choices_to={"subject_type": SubjectType.EVENT},
        related_name="+",
        verbose_name=_("event types"),
    )

    class Meta:
        app_label = "leprikon"

    def copy_relations(self, oldinstance):
        self.event_types.set(oldinstance.event_types.all())

    @cached_property
    def all_event_types(self):
        return list(self.event_types.all())

    def render(self, context):
        school_year = (
            self.school_year or getattr(context.get("request"), "school_year") or SchoolYear.objects.get_current()
        )

        from ..forms.subjects import SubjectFilterForm

        form = SubjectFilterForm(
            subject_type_type=SubjectType.EVENT,
            subject_types=self.all_event_types,
            school_year=school_year,
            is_staff=context["request"].user.is_staff,
            data=context["request"].GET,
        )
        context.update(
            {
                "school_year": school_year,
                "form": form,
                "events": form.get_queryset(),
            }
        )
        return context



from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class StartEndMixin(object):
    def validate(self):
        if self.start and self.end and self.start > self.end:
            raise ValidationError(
                {
                    "start": [_("Start must be before end")],
                    "end": [_("End must be later than start")],
                }
            )

    def clean_fields(self, exclude=None):
        super().clean_fields(exclude)
        self.validate()

    def save(self, *args, **kwargs):
        self.validate()
        super().save(*args, **kwargs)



from adminsortable2.admin import SortableInlineAdminMixin
from django import forms
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.shortcuts import render
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..models.roles import Contact, Leader
from ..models.schoolyear import SchoolYear
from ..models.subjects import SubjectType
from ..utils import attributes
from .export import AdminExportMixin
from .filters import SchoolYearListFilter
from .messages import SendMessageAdminMixin


class ContactInlineAdmin(SortableInlineAdminMixin, admin.TabularInline):
    model = Contact
    extra = 0


@admin.register(Leader)
class LeaderAdmin(AdminExportMixin, SendMessageAdminMixin, admin.ModelAdmin):
    actions = SendMessageAdminMixin.actions + AdminExportMixin.actions + ("add_school_year",)
    filter_horizontal = ("school_years",)
    inlines = (ContactInlineAdmin,)
    search_fields = ("user__first_name", "user__last_name", "contacts__contact")
    list_display = (
        "id",
        "first_name",
        "last_name",
        "email",
        "courses_link",
        "events_link",
        "orderables_link",
        "contacts",
        "user_link",
        "icon",
    )
    list_export = (
        "id",
        "user__first_name",
        "user__last_name",
        "user__email",
        "contacts__contact_type",
        "contacts__contact",
        "contacts__public",
    )
    ordering = ("user__first_name", "user__last_name")
    list_filter = (("school_years", SchoolYearListFilter),)
    raw_id_fields = ("user",)

    @attributes(short_description=_("Add selected leaders to another school year"))
    def add_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_("Target school year"),
                help_text=_("All selected leaders will be added to selected school year."),
                queryset=SchoolYear.objects.all(),
            )

        if request.POST.get("post", "no") == "yes":
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data["school_year"]
                for leader in queryset.all():
                    leader.school_years.add(school_year)
                self.message_user(request, _("Selected leaders were added to school year {}.").format(school_year))
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Select target school year"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="add_school_year",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )

    @attributes(admin_order_field="user__first_name", short_description=_("first name"))
    def first_name(self, obj):
        return obj.user.first_name

    @attributes(admin_order_field="user__last_name", short_description=_("last name"))
    def last_name(self, obj):
        return obj.user.last_name

    @attributes(admin_order_field="user__email", short_description=_("email"))
    def email(self, obj):
        return obj.user.email

    @attributes(short_description=_("contacts"))
    def contacts(self, obj):
        return ", ".join(c.contact for c in obj.all_contacts)

    @attributes(short_description=_("user"))
    def user_link(self, obj):
        return mark_safe(
            '<a href="{url}">{user}</a>'.format(
                url=reverse(
                    f"admin:{obj.user._meta.app_label}_{obj.user._meta.model_name}_change", args=(obj.user.id,)
                ),
                user=obj.user,
            )
        )

    @cached_property
    def courses_url(self):
        return reverse("admin:leprikon_course_changelist")

    @cached_property
    def events_url(self):
        return reverse("admin:leprikon_event_changelist")

    @cached_property
    def orderables_url(self):
        return reverse("admin:leprikon_orderable_changelist")

    @attributes(short_description=_("courses"))
    def courses_link(self, obj):
        return mark_safe(
            '<a href="{url}?leaders__id={leader}">{count}</a>'.format(
                url=self.courses_url,
                leader=obj.id,
                count=obj.subjects.filter(subject_type__subject_type=SubjectType.COURSE).count(),
            )
        )

    @attributes(short_description=_("events"))
    def events_link(self, obj):
        return mark_safe(
            '<a href="{url}?leaders__id={leader}">{count}</a>'.format(
                url=self.events_url,
                leader=obj.id,
                count=obj.subjects.filter(subject_type__subject_type=SubjectType.EVENT).count(),
            )
        )

    @attributes(short_description=_("orderable events"))
    def orderables_link(self, obj):
        return mark_safe(
            '<a href="{url}?leaders__id={leader}">{count}</a>'.format(
                url=self.orderables_url,
                leader=obj.id,
                count=obj.subjects.filter(subject_type__subject_type=SubjectType.ORDERABLE).count(),
            )
        )

    @attributes(short_description=_("photo"))
    def icon(self, obj):
        try:
            return mark_safe('<img src="{}" alt="{}"/>'.format(obj.photo.icons["48"], obj.photo.label))
        except (AttributeError, KeyError):
            return ""

    def get_message_recipients(self, request, queryset):
        return (
            get_user_model()
            .objects.filter(
                leprikon_leader__in=queryset,
            )
            .distinct()
        )



from adminsortable2.admin import SortableAdminMixin
from django import forms
from django.contrib import admin
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..forms.timesheets import TimesheetEntryAdminForm
from ..models.timesheets import Timesheet, TimesheetEntry, TimesheetEntryType
from ..utils import attributes
from .export import AdminExportMixin
from .filters import LeaderListFilter
from .journals import JournalLeaderEntryInlineAdmin


@admin.register(TimesheetEntry)
class TimesheetEntryAdmin(AdminExportMixin, admin.ModelAdmin):
    form = TimesheetEntryAdminForm
    date_hierarchy = "date"
    list_display = ("timesheet", "date", "start", "end", "duration", "entry_type")
    list_filter = ("entry_type", ("timesheet__leader", LeaderListFilter))
    ordering = (
        "-date",
        "-start",
    )

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.timesheet.submitted:
            return ("timesheet", "entry_type", "date", "start", "end")
        return ()

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if obj and obj.timesheet.submitted:
            return False
        return super().has_delete_permission(request, obj)


class TimesheetEntryInlineAdmin(admin.TabularInline):
    class form(forms.ModelForm):
        class Meta:
            model = TimesheetEntry
            fields = []

    model = TimesheetEntry
    ordering = ("date", "start")
    readonly_fields = ("date", "start", "end", "entry_type", "description_html", "edit_link")

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        if obj and obj.submitted:
            return False
        return super().has_delete_permission(request, obj)

    @attributes(admin_order_field="description", short_description=_("description"))
    def description_html(self, obj):
        return mark_safe(obj.description)

    @attributes(short_description="")
    def edit_link(self, obj):
        return mark_safe(
            '<a href="{url}" title="{title}" target="_blank">{edit}</a>'.format(
                url=reverse("admin:leprikon_timesheetentry_change", args=[obj.id]),
                title=_("update entry"),
                edit=_("edit"),
            )
        )


@admin.register(Timesheet)
class TimesheetAdmin(AdminExportMixin, admin.ModelAdmin):
    list_display = ("leader", "period", "group_durations", "submitted", "paid")
    list_filter = (("leader", LeaderListFilter), "period")
    inlines = (TimesheetEntryInlineAdmin, JournalLeaderEntryInlineAdmin)
    actions = AdminExportMixin.actions + ("submit", "set_paid")

    # do not allow to add timesheets in admin
    # timesheets are created automatically
    def has_add_permission(self, request, obj=None):
        return False

    # do not allow to delete submitted timesheets
    def has_delete_permission(self, request, obj=None):
        return super().has_delete_permission(request, obj) and (obj is None or not obj.submitted)

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:

            def delete_selected(model_admin, request, queryset):
                queryset = queryset.filter(submitted=False)
                return admin.actions.delete_selected(model_admin, request, queryset)

            actions["delete_selected"] = (delete_selected, *actions["delete_selected"][1:])
        return actions

    @attributes(short_description=_("group durations"))
    def group_durations(self, obj):
        return mark_safe(
            "<br/>".join(
                "<b>{name}</b>: {duration}".format(
                    name=group.name,
                    duration=group.duration,
                )
                for group in obj.groups
            )
        )

    @attributes(short_description=_("Mark selected timesheets as paid"))
    def set_paid(self, request, queryset):
        queryset.update(submitted=True, paid=True)
        self.message_user(request, _("Selected timesheets where marked as paid."))

    @attributes(short_description=_("Submit selected timesheets"))
    def submit(self, request, queryset):
        queryset.update(submitted=True)
        self.message_user(request, _("Selected timesheets where submitted."))


@admin.register(TimesheetEntryType)
class TimesheetEntryTypeAdmin(SortableAdminMixin, admin.ModelAdmin):
    list_display = ("name",)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2022-02-20 22:45
from __future__ import unicode_literals
from itertools import chain
from json import dumps, loads

from string import ascii_lowercase

from django.db import migrations, models
from django.utils.crypto import get_random_string
from django.utils.text import slugify
import django.db.models.deletion


def set_question_slug(apps, schema_editor):
    Question = apps.get_model("leprikon", "Question")
    SubjectRegistrationParticipant = apps.get_model("leprikon", "SubjectRegistrationParticipant")
    Participant = apps.get_model("leprikon", "SubjectRegistrationParticipant")

    new_keys = {}
    for question in Question.objects.all():
        question.slug = slugify(question.name)[:50]
        while Question.objects.filter(slug=question.slug).exists():
            slug_base = slugify(question.name)[:45]
            rand_string = get_random_string(4, ascii_lowercase)
            question.slug = f"{slug_base}-{rand_string}"
        question.save()
        new_keys[question.name] = question.slug

    for Model in (SubjectRegistrationParticipant, Participant):
        for obj in Model.objects.all():
            obj.answers = dumps({new_keys[key]: value for key, value in loads(obj.answers).items() if key in new_keys})
            obj.save()


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0068_participant_answers"),
    ]

    operations = [
        migrations.AddField(
            model_name="question",
            name="slug",
            field=models.SlugField(null=True, verbose_name="unique identifier"),
        ),
        migrations.RunPython(set_question_slug),
        migrations.AlterModelOptions(
            name="question",
            options={
                "ordering": ("name",),
                "verbose_name": "additional question",
                "verbose_name_plural": "additional questions",
            },
        ),
        migrations.AlterField(
            model_name="question",
            name="name",
            field=models.CharField(max_length=150, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="question",
            name="question",
            field=models.CharField(max_length=150, verbose_name="question"),
        ),
    ]



"""
Django settings for leprikon.site project.

Generated by 'django-admin startproject' using Django 3.2.15.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""

import os
import re

from django.utils.translation import gettext_lazy as _

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.environ.get("BASE_DIR", os.getcwd())
DATA_DIR = os.environ.get("DATA_DIR", os.path.join(BASE_DIR, "data"))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get("SECRET_KEY", os.environ.get("DEBUG"))
if not SECRET_KEY:
    from string import ascii_letters, digits, punctuation

    from django.utils.crypto import get_random_string

    SECRET_KEY = get_random_string(
        64,
        ascii_letters + digits + punctuation,
    )

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get("DEBUG", "") and True or False
DBDEBUG = "DB" in os.environ.get("DEBUG", "").split(",")
DEBUG_TEMPLATE = "TEMPLATE" in os.environ.get("DEBUG", "").split(",")
DEBUG_LML = "LML" in os.environ.get("DEBUG", "").split(",")

ALLOWED_HOSTS = [h.strip() for h in os.environ.get("ALLOWED_HOSTS", "*").split(",")]


# Application definition

INSTALLED_APPS = [
    "user_unique_email",
    "leprikon.site",
    "leprikon",
    "adminsortable2",
    "bankreader",
    "djangocms_admin_style",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sites",
    "django.contrib.sitemaps",
    "django_cron",
    "django_pays",
    "djangocms_audio",
    "djangocms_file",
    "djangocms_picture",
    "djangocms_video",
    "cms",
    "haystack",
    "mathfilters",
    "menus",
    "multiselectfield",
    "sekizai",
    "template_admin",
    "treebeard",
    "filer",
    "easy_thumbnails",
    "djangocms_link",
    "djangocms_text_ckeditor",
    "ganalytics",
    "html2rml",
    "qr_code",
    "social_django",
    "verified_email_field",
    "drf_spectacular",
    "drf_spectacular_sidecar",
    "rest_framework",
]

CMSPLUGIN_FILER_MIGRATED = os.environ.get("CMSPLUGIN_FILER_MIGRATED", "no").lower() in ("yes", "true")
if not CMSPLUGIN_FILER_MIGRATED:
    print(
        """
        Please migrate all content using plugins from cmsplugin_filer
        to corresponding djangocms_* plugins.
        When done, run commands:

        for app in cmsplugin_filer_{folder,file,image,link,teaser,video}; do
            leprikon migrate $app zero
        done

        and set CMSPLUGIN_FILER_MIGRATED environment variable to "yes".
        """
    )
    INSTALLED_APPS += [
        "cmsplugin_filer_file",
        "cmsplugin_filer_folder",
        "cmsplugin_filer_image",
        "cmsplugin_filer_link",
        "cmsplugin_filer_teaser",
        "cmsplugin_filer_utils",
        "cmsplugin_filer_video",
    ]

MIDDLEWARE = [
    "cms.middleware.utils.ApphookReloadMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "cms.middleware.user.CurrentUserMiddleware",
    "cms.middleware.page.CurrentPageMiddleware",
    "cms.middleware.toolbar.ToolbarMiddleware",
    "cms.middleware.language.LanguageCookieMiddleware",
    "social_django.middleware.SocialAuthExceptionMiddleware",
    "leprikon.middleware.LeprikonMiddleware",
]

CRON_CLASSES = [
    "leprikon.cronjobs.SendPaymentRequest",
]

CRON_SEND_PAYMENT_REQUEST_TIME = os.environ.get("CRON_SEND_PAYMENT_REQUEST_TIME", "8:00")

ROOT_URLCONF = os.environ.get("ROOT_URLCONF", "leprikon.site.urls")

BASE_TEMPLATE_LOADERS = [
    "django.template.loaders.filesystem.Loader",
    "django.template.loaders.app_directories.Loader",
]
TEMPLATE_LOADERS = [
    (
        "template_admin.loader.Loader",
        BASE_TEMPLATE_LOADERS
        if DEBUG_TEMPLATE
        else [
            ("django.template.loaders.cached.Loader", BASE_TEMPLATE_LOADERS),
        ],
    ),
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.csrf",
                "django.template.context_processors.tz",
                "django.template.context_processors.static",
                "sekizai.context_processors.sekizai",
                "cms.context_processors.cms_settings",
                "social_django.context_processors.backends",
                "social_django.context_processors.login_redirect",
            ],
            "debug": DEBUG_TEMPLATE,
            "loaders": TEMPLATE_LOADERS,
        },
    },
]

WSGI_APPLICATION = os.environ.get("UWSGI_MODULE", "leprikon.site.wsgi:application").replace(":", ".")

# Social configuration
SOCIAL_AUTH_FACEBOOK_SCOPE = ["email"]
SOCIAL_AUTH_FACEBOOK_PROFILE_EXTRA_PARAMS = {"fields": "id,name,email"}
SOCIAL_AUTH_FACEBOOK_KEY = os.environ.get("SOCIAL_AUTH_FACEBOOK_KEY")
SOCIAL_AUTH_FACEBOOK_SECRET = os.environ.get("SOCIAL_AUTH_FACEBOOK_SECRET")

SOCIAL_AUTH_GOOGLE_OAUTH2_SCOPE = ["profile", "email"]
SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = os.environ.get("SOCIAL_AUTH_GOOGLE_OAUTH2_KEY")
SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = os.environ.get("SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET")

SOCIAL_AUTH_GOOGLE_PLUS_SCOPE = ["profile", "email"]
SOCIAL_AUTH_GOOGLE_PLUS_KEY = os.environ.get("SOCIAL_AUTH_GOOGLE_PLUS_KEY")
SOCIAL_AUTH_GOOGLE_PLUS_SECRET = os.environ.get("SOCIAL_AUTH_GOOGLE_PLUS_SECRET")

SOCIAL_AUTH_PIPELINE = (
    "social_core.pipeline.social_auth.social_details",
    "social_core.pipeline.social_auth.social_uid",
    "social_core.pipeline.social_auth.social_user",
    "social_core.pipeline.user.get_username",
    "social_core.pipeline.social_auth.associate_by_email",
    "social_core.pipeline.user.create_user",
    "social_core.pipeline.social_auth.associate_user",
    "social_core.pipeline.social_auth.load_extra_data",
    "social_core.pipeline.user.user_details",
)

# Authentication configuration
AUTHENTICATION_BACKENDS = (
    (["social_core.backends.facebook.FacebookOAuth2"] if SOCIAL_AUTH_FACEBOOK_KEY else [])
    + (["social_core.backends.google.GoogleOAuth2"] if SOCIAL_AUTH_GOOGLE_OAUTH2_KEY else [])
    + (["social_core.backends.google.GooglePlusAuth"] if SOCIAL_AUTH_GOOGLE_PLUS_KEY else [])
    + [
        "verified_email_field.auth.VerifiedEmailBackend",
        "django.contrib.auth.backends.ModelBackend",
    ]
)

# Database
# https://docs.djangoproject.com/en/3.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": os.environ.get("DATABASE_ENGINE", "django.db.backends.sqlite3"),
        "NAME": os.environ.get("DATABASE_NAME", os.path.join(DATA_DIR, "db.sqlite3")),
        "HOST": os.environ.get("DATABASE_HOST", None),
        "PORT": os.environ.get("DATABASE_PORT", None),
        "USER": os.environ.get("DATABASE_USER", None),
        "PASSWORD": os.environ.get("DATABASE_PASSWORD", None),
    }
}

if DATABASES["default"]["ENGINE"].endswith("mysql"):
    DATABASES["default"]["OPTIONS"] = {
        "charset": "utf8mb4",
        "use_unicode": True,
        "init_command": (
            "ALTER DATABASE `{name}` CHARACTER SET utf8mb4 COLLATE {collate}; "
            "SET default_storage_engine=INNODB; SET sql_mode=STRICT_TRANS_TABLES; ".format(
                name=DATABASES["default"]["NAME"],
                collate=os.environ.get("DATABASE_COLLATE", "utf8mb4_czech_ci"),
            )
        ),
    }

# Leprikon URL
LEPRIKON_DOMAIN = os.environ.get("LEPRIKON_DOMAIN")
LEPRIKON_URL = os.environ.get("LEPRIKON_URL", LEPRIKON_DOMAIN and f"https://{LEPRIKON_DOMAIN}")

# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = (
    [
        {
            "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
        },
        {
            "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
        },
    ]
    if not DEBUG
    else []
)

# Custom User model
AUTH_USER_MODEL = "user_unique_email.User"

# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGES = (("cs", _("Czech")),)

LANGUAGE_CODE = LANGUAGES[0][0]

TIME_ZONE = "Europe/Prague"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# set max upload size to 10MB
DATA_UPLOAD_MAX_MEMORY_SIZE = 10 * 1024 * 1024
DATA_UPLOAD_MAX_NUMBER_FIELDS = 10000

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"
MEDIA_URL = "/media/"
MEDIA_ROOT = os.path.join(BASE_DIR, "htdocs", "media")
STATIC_ROOT = os.path.join(BASE_DIR, "htdocs", "static")

STATICFILES_DIRS = [os.path.join(BASE_DIR, "static")] if os.path.exists(os.path.join(BASE_DIR, "static")) else []
STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
    "staticfiles_downloader.DownloaderFinder",
]
if not DEBUG:
    STATICFILES_STORAGE = "leprikon.staticfiles.NotStrictManifestStaticFilesStorage"

LOCALE_PATHS = [os.path.join(BASE_DIR, "locale")] if os.path.exists(os.path.join(BASE_DIR, "locale")) else []

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

# Logging
# https://docs.djangoproject.com/en/3.2/topics/logging/

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {"format": "%(levelname)s %(asctime)s %(module)s " "%(process)d %(thread)d %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
            "filters": [],
        },
    },
    "loggers": {
        "": {
            "level": "DEBUG" if DEBUG else "WARNING",
            "handlers": ["console"],
        },
        "django.db.backends": {
            "level": "DEBUG" if DBDEBUG else "ERROR",
            "propagate": True,
        },
        "parso": {
            "level": "INFO",
        },
    },
}
if not DEBUG_LML:
    LOGGING["loggers"]["lml"] = {"level": "WARNING"}
    LOGGING["loggers"]["pyexcel"] = {"level": "WARNING"}
    LOGGING["loggers"]["pyexcel_io"] = {"level": "WARNING"}


# Sentry configuration
if not DEBUG and os.environ.get("SENTRY_DSN"):
    import sentry_sdk
    from sentry_sdk.integrations.django import DjangoIntegration

    from leprikon import __version__

    sentry_sdk.init(
        dsn=os.environ.get("SENTRY_DSN"),
        integrations=[DjangoIntegration()],
        send_default_pii=True,
        release=__version__,
    )


# Caching and session configuration
# Only configure remote cache if explicitly specified in the environment.
# Otherwise it would be enabled during `leprikon collectstatic` during build.
if "CACHE_LOCATION" in os.environ:
    CACHES = {
        "default": {
            "BACKEND": os.environ.get("CACHE_BACKEND", "django_redis.cache.RedisCache"),
            "LOCATION": os.environ.get("CACHE_LOCATION"),
        }
    }
    if CACHES["default"]["BACKEND"] == "django_redis.cache.RedisCache":
        CACHES["default"]["OPTIONS"] = {
            "CLIENT_CLASS": "django_redis.client.DefaultClient",
        }
    if "CACHE_KEY_PREFIX" in os.environ:
        CACHES["default"]["KEY_PREFIX"] = os.environ["CACHE_KEY_PREFIX"]

SESSION_ENGINE = "django.contrib.sessions.backends.cache"
SESSION_CACHE_ALIAS = "default"
if "SESSION_COOKIE_AGE" in os.environ:
    SESSION_COOKIE_AGE = int(os.environ["SESSION_COOKIE_AGE"])
if "SESSION_STAFF_COOKIE_AGE" in os.environ:
    SESSION_STAFF_COOKIE_AGE = int(os.environ["SESSION_STAFF_COOKIE_AGE"])

# Login / logout urls
LOGIN_URL = "leprikon:user_login"
LOGOUT_URL = "leprikon:user_logout"
LOGIN_REDIRECT_URL = "leprikon:summary"
LOGIN_ERROR_URL = LOGIN_URL

# Email configuration
SERVER_EMAIL = os.environ.get("SERVER_EMAIL", "leprikon@localhost")
SERVER_EMAIL_PLAIN = re.search("[^<]+@[^>]+", SERVER_EMAIL).group()

EMAIL_BACKEND = (
    "django.core.mail.backends.console.EmailBackend"
    if "EMAIL" in os.environ.get("DEBUG", "").split(",")
    else "django.core.mail.backends.smtp.EmailBackend"
)
EMAIL_HOST = os.environ.get("EMAIL_HOST", "localhost")
EMAIL_PORT = int(os.environ.get("EMAIL_PORT", "25"))
EMAIL_HOST_USER = os.environ.get("EMAIL_HOST_USER", "")
EMAIL_HOST_PASSWORD = os.environ.get("EMAIL_HOST_PASSWORD", "")
EMAIL_SUBJECT_PREFIX = os.environ.get("EMAIL_SUBJECT_PREFIX", "[Leprikón] ").strip() + " "
EMAIL_USE_SSL = os.environ.get("EMAIL_USE_SSL", "").lower() in ("y", "yes", "t", "true", "1", "on")
DEFAULT_FROM_EMAIL = SERVER_EMAIL

# CMS configuration
# http://djangocms.readthedocs.io/en/latest/reference/configuration/
SITE_ID = 1

CMS_LANGUAGES = {
    SITE_ID: [{"code": lang[0], "name": lang[1]} for lang in LANGUAGES],
    "default": {
        "public": True,
        "fallbacks": [LANGUAGE_CODE],
        "hide_untranslated": True,
        "redirect_on_fallback": True,
    },
}

CMS_TEMPLATES = [
    ("leprikon/cms.html", "Leprikon"),
]

CMS_PERMISSION = True

CMS_PLACEHOLDER_CONF = {}

THUMBNAIL_DEBUG = "THUMBNAIL" in os.environ.get("DEBUG", "").split(",")

THUMBNAIL_PROCESSORS = (
    "easy_thumbnails.processors.colorspace",
    "easy_thumbnails.processors.autocrop",
    "filer.thumbnail_processors.scale_and_crop_with_subject_location",
    "easy_thumbnails.processors.filters",
)

# Haystack configuration
HAYSTACK_CONNECTIONS = {
    "default": {"ENGINE": "haystack.backends.whoosh_backend.WhooshEngine"},
}
for key, value in os.environ.items():
    if key.startswith("HAYSTACK_"):
        HAYSTACK_CONNECTIONS["default"][key[len("HAYSTACK_") :]] = value
if HAYSTACK_CONNECTIONS["default"]["ENGINE"] == "haystack.backends.whoosh_backend.WhooshEngine":
    HAYSTACK_CONNECTIONS["default"].setdefault("PATH", os.path.join(DATA_DIR, "whoosh_index"))

# Google Analytics configuration
GANALYTICS_TRACKING_CODE = os.environ.get("GANALYTICS_TRACKING_CODE")

# Leprikon configuration
PRICE_DECIMAL_PLACES = 0
COUNTRIES_FIRST = ["CZ", "SK"]
if "LEPRIKON_VARIABLE_SYMBOL_EXPRESSION" in os.environ:
    LEPRIKON_VARIABLE_SYMBOL_EXPRESSION = os.environ.get("LEPRIKON_VARIABLE_SYMBOL_EXPRESSION")

LEPRIKON_SHOW_SUBJECT_CODE = os.environ.get("LEPRIKON_SHOW_SUBJECT_CODE", "").lower() in ("1", "y", "yes", "t", "true")

X_FRAME_OPTIONS = "SAMEORIGIN"


#
# Rest Framework
#
REST_FRAMEWORK = {
    "DEFAULT_PARSER_CLASSES": [
        "djangorestframework_camel_case.parser.CamelCaseJSONParser",
    ],
    "DEFAULT_RENDERER_CLASSES": [
        "djangorestframework_camel_case.render.CamelCaseJSONRenderer",
        "djangorestframework_camel_case.render.CamelCaseBrowsableAPIRenderer",
    ],
    "DEFAULT_SCHEMA_CLASS": "leprikon.api.openapi.Schema",
}

SPECTACULAR_SETTINGS = {
    # header
    "TITLE": "Leprikon API",
    "DESCRIPTION": "",
    "CONTACT": None,
    "LICENSE": None,
    "TOS": None,
    "VERSION": "v1",
    "SERVERS": [
        {"url": "http://localhost:8000/", "description": "local"},
    ],
    # options
    "CAMELIZE_NAMES": True,
    "POSTPROCESSING_HOOKS": [
        "drf_spectacular.contrib.djangorestframework_camel_case.camelize_serializer_fields",
        "drf_spectacular.hooks.postprocess_schema_enums",
    ],
    "REDOC_DIST": "SIDECAR",
    "SCHEMA_PATH_PREFIX": "/api",
    "SERVE_INCLUDE_SCHEMA": False,
    "SWAGGER_UI_DIST": "SIDECAR",
    "SWAGGER_UI_FAVICON_HREF": "SIDECAR",
}



from django.conf import settings
from django.contrib import admin
from django.http import HttpRequest, HttpResponseRedirect
from django.shortcuts import redirect

from ..models.leprikonsite import LeprikonSite


@admin.register(LeprikonSite)
class LeprikonSiteAdmin(admin.ModelAdmin):
    list_display = ("name",)
    filter_horizontal = ("registration_agreements",)

    def changelist_view(self, request: HttpRequest, extra_context=None) -> HttpResponseRedirect:
        return redirect("admin:leprikon_leprikonsite_change", object_id=settings.SITE_ID)

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls.base import reverse_lazy as reverse
from django.utils.translation import gettext_lazy as _

from ..forms.journals import JournalCreateForm, JournalEntryForm, JournalLeaderEntryForm, JournalUpdateForm
from ..models.journals import Journal, JournalEntry, JournalLeaderEntry, Subject
from ..utils import reverse_with_back
from .generic import CreateView, DeleteView, DetailView, TemplateView, UpdateView


class AlternatingView(TemplateView):
    template_name = "leprikon/alternating.html"

    def get_title(self):
        return _("Alternating in school year {}").format(self.request.school_year)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["alternate_leader_entries"] = self.request.leader.get_alternate_leader_entries(self.request.school_year)
        return context


class JournalQuerySetMixin:
    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leaders=self.request.leader)
        return qs


class JournalView(JournalQuerySetMixin, DetailView):
    model = Journal
    template_name_suffix = "_journal"


class JournalCreateView(CreateView):
    model = Journal
    form_class = JournalCreateForm
    template_name = "leprikon/journal_form.html"
    title = _("New journal")

    def dispatch(self, request, subject):
        kwargs = {"id": subject}
        if not self.request.user.is_staff:
            kwargs["leaders"] = self.request.leader
        self.subject = get_object_or_404(Subject, **kwargs)
        return super().dispatch(request)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["subject"] = self.subject
        return kwargs

    def get_message(self):
        return _("New journal {} has been created.").format(self.object)

    def get_success_url(self):
        return reverse_with_back(self.request, "leprikon:journal_update", args=(self.object.id,))


class JournalUpdateView(JournalQuerySetMixin, UpdateView):
    model = Journal
    form_class = JournalUpdateForm
    success_url = reverse("leprikon:summary")
    template_name = "leprikon/journal_form.html"
    title = _("Change journal")


class JournalDeleteView(DeleteView):
    model = Journal
    title = _("Delete journal")
    message = _("Journal has been deleted.")

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(subject__leaders=self.request.leader)
        return qs

    def get_object(self):
        obj = super().get_object()
        if obj.all_journal_entries:
            raise Http404()
        return obj

    def get_question(self):
        return _("Do You really want to delete the journal {}?").format(self.object)


class JournalEntryCreateView(CreateView):
    model = JournalEntry
    form_class = JournalEntryForm
    template_name = "leprikon/journalentry_form.html"
    title = _("New journal entry")
    message = _("The journal entry has been created.")

    def dispatch(self, request, *args, **kwargs):
        if self.request.user.is_staff:
            self.journal = get_object_or_404(Journal, id=int(kwargs.pop("journal")))
        else:
            self.journal = get_object_or_404(Journal, id=int(kwargs.pop("journal")), leaders=self.request.leader)
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["journal"] = self.journal
        return kwargs


class JournalEntryUpdateView(UpdateView):
    model = JournalEntry
    form_class = JournalEntryForm
    template_name = "leprikon/journalentry_form.html"
    title = _("Change journal entry")
    message = _("The journal entry has been updated.")

    def get_object(self):
        obj = super().get_object()
        if self.request.user.is_staff or self.request.leader in obj.journal.all_leaders + obj.all_alternates:
            return obj
        else:
            raise Http404()


class JournalEntryDeleteView(DeleteView):
    model = JournalEntry
    title = _("Delete journal entry")
    message = _("The journal entry has been deleted.")

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(journal__leaders=self.request.leader)
        return qs

    def get_object(self):
        obj = super().get_object()
        if obj.affects_submitted_timesheets:
            raise Http404()
        return obj

    def get_question(self):
        return _("Do You really want to delete journal entry?")


class JournalLeaderEntryUpdateView(UpdateView):
    model = JournalLeaderEntry
    form_class = JournalLeaderEntryForm
    template_name = "leprikon/journalleaderentry_form.html"
    title = _("Change timesheet entry")
    message = _("The timesheet entry has been updated.")

    def get_object(self):
        obj = super().get_object()
        if (
            self.request.user.is_staff
            or obj.timesheet.leader == self.request.leader
            or self.request.leader in obj.journal_entry.journal.all_leaders
        ):
            return obj
        else:
            raise Http404()


class JournalLeaderEntryDeleteView(DeleteView):
    model = JournalLeaderEntry
    title = _("Delete timesheet entry")
    message = _("The timesheet entry has been deleted.")

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                timesheet__leader=self.request.leader,
                timesheet__submitted=False,
            )
        )

    def get_question(self):
        return _("Do You really want to delete timesheet entry?")



from django.db import models
from django.utils.translation import gettext_lazy as _


class Citizenship(models.Model):
    name = models.CharField(_("name"), max_length=50)
    order = models.IntegerField(_("order"), blank=True, default=0)
    require_birth_num = models.BooleanField(_("require birth number"), default=True)

    class Meta:
        app_label = "leprikon"
        ordering = ("order",)
        verbose_name = _("citizenship")
        verbose_name_plural = _("citizenships")

    def __str__(self):
        return self.name



from django.http import Http404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from ..forms.timesheets import TimesheetEntryForm
from ..models.timesheets import Timesheet, TimesheetEntry
from .generic import ConfirmUpdateView, CreateView, DeleteView, DetailView, ListView, UpdateView


class TimesheetListView(ListView):
    model = Timesheet
    preview_template = "leprikon/timesheet_preview.html"
    paginate_by = 6
    add_label = _("add entry")
    add_url = reverse_lazy("leprikon:timesheetentry_create")

    def get_title(self):
        return _("Timesheets")

    def get_queryset(self):
        return super().get_queryset().filter(leader=self.request.leader)


class TimesheetDetailView(DetailView):
    model = Timesheet

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(leader=self.request.leader)
        return qs


class TimesheetSubmitView(ConfirmUpdateView):
    model = Timesheet
    title = _("Submit timesheet")

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                leader=self.request.leader,
                submitted=False,
            )
        )

    def get_question(self):
        return _(
            "Are you sure You want to submit the timesheet now? "
            "You won't be able to edit the entries for {} any more."
        ).format(self.object.period.name)

    def get_message(self):
        return _("The timesheet for () has been submitted.").format(self.object.period.name)

    def confirmed(self):
        self.object.submitted = True
        self.object.save()


class TimesheetEntryCreateView(CreateView):
    model = TimesheetEntry
    form_class = TimesheetEntryForm
    template_name = "leprikon/timesheetentry_form.html"
    title = _("New timesheet entry")
    message = _("New timesheet entry has been created.")

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["leader"] = self.request.leader
        return kwargs


class TimesheetEntryUpdateView(UpdateView):
    model = TimesheetEntry
    form_class = TimesheetEntryForm
    template_name = "leprikon/timesheetentry_form.html"
    title = _("Change timesheet entry")
    message = _("The timesheet entry has been updated.")

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(timesheet__leader=self.request.leader)
        return qs


class TimesheetEntryDeleteView(DeleteView):
    model = TimesheetEntry
    title = _("Delete timesheet entry")
    message = _("The timesheet entry has been deleted.")

    def get_queryset(self):
        qs = super().get_queryset()
        if not self.request.user.is_staff:
            qs = qs.filter(timesheet__leader=self.request.leader)
        return qs

    def get_object(self):
        obj = super().get_object()
        if obj.timesheet.submitted:
            raise Http404()
        return obj

    def get_question(self):
        return _("Do You really want to delete timesheet entry?")
